import os
import multiprocessing
import threading
import ctypes
import datetime
import zlib
import pickle

from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, OrderedDict
from multiprocessing.managers import SyncManager
from queue import Queue



from ms_deisotope.output.common import LCMSMSQueryInterfaceMixin
from ms_deisotope.data_source import ProcessedRandomAccessScanSource

from glycresoft import serialize
from glycresoft.serialize.hypothesis.hypothesis import GlycopeptideHypothesis


from glycresoft.tandem.spectrum_match.solution_set import MultiScoreSpectrumSolutionSet
from glycresoft.tandem.spectrum_match.spectrum_match import MultiScoreSpectrumMatch, SpectrumMatch

from glycresoft.task import (
    TaskBase, Pipeline,
    TaskExecutionSequence,
    make_shared_memory_manager,
    IPCLoggingManager, LoggingHandlerToken)

from glycresoft.chromatogram_tree import Unmodified, MassShift

from glycresoft.structure import ScanStub

from glycresoft.database import disk_backed_database, mass_collection
from glycresoft.structure.structure_loader import PeptideDatabaseRecord
from glycresoft.structure.scan import ScanInformationLoader
from glycresoft.composition_distribution_model.site_model import (
    GlycoproteomePriorAnnotator)

from glycresoft.tandem.oxonium_ions import OxoniumIndex, SignatureIonIndex, single_signatures, compound_signatures

from glycresoft.chromatogram_tree.chromatogram import Chromatogram, GlycopeptideChromatogram
from ...chromatogram_mapping import (
    ChromatogramMSMSMapper,
    TandemAnnotatedChromatogram,
    TandemSolutionsWithoutChromatogram
)
from ...temp_store import TempFileManager
from ...spectrum_evaluation import group_by_precursor_mass
from ...spectrum_match import SpectrumMatchClassification
from ...workflow import SearchEngineBase

from ..scoring import LogIntensityScorer, GlycopeptideSpectrumMatcherBase

from .journal import (
    JournalFileWriter,
    JournalFileReader,
    JournalingConsumer,
    SolutionSetGrouper)

from .search_space import (
    PeptideGlycosylator,
    PredictiveGlycopeptideSearch,
    GlycanCombinationRecord,
    StructureClassification)

from .searcher import (
    SpectrumBatcher,
    BatchMapper,
    SemaphoreBoundMatcherExecutor,
    SemaphoreBoundMapperExecutor)

from .multipart_fdr import GlycopeptideFDREstimator, GlycopeptideFDREstimationStrategy


def make_memory_database_proxy_resolver(path, n_glycan=None, o_glycan=None, hypothesis_id=1):
    if n_glycan is None or o_glycan is None:
        config = _determine_database_contents(path, hypothesis_id)
        if o_glycan is None:
            o_glycan = config['o_glycan']
        if n_glycan is None:
            n_glycan = config['n_glycan']
    loader = PeptideDatabaseProxyLoader(path, n_glycan, o_glycan, hypothesis_id)
    # Make it possible to load the session here without loading the peptides
    proxy = mass_collection.MassCollectionProxy(
        loader, loader.session_resolver, hypothesis_id, loader.hypothesis_resolver)
    return proxy


def make_disk_backed_peptide_database(path, hypothesis_id=1, **kwargs):
    peptide_db = disk_backed_database.PeptideDiskBackedStructureDatabase(
        path, cache_size=100, hypothesis_id=hypothesis_id)
    peptide_db = mass_collection.TransformingMassCollectionAdapter(
        peptide_db, PeptideDatabaseRecord.from_record)
    return peptide_db


class FetchManyIterator(object):
    def __init__(self, cursor, batch_size=1000000):
        self.cursor = cursor
        self.batch_size = batch_size

    def __iter__(self):
        result_set = self.cursor.fetchmany(self.batch_size)
        while result_set:
            for x in result_set:
                yield x
            result_set = self.cursor.fetchmany(self.batch_size)


class PeptideDatabaseProxyLoader(TaskBase):
    path: os.PathLike
    n_glycan: bool
    o_glycan: bool
    hypothesis_id: int

    def __init__(self, path, n_glycan=True, o_glycan=True, hypothesis_id=1):
        self.path = path
        self.n_glycan = n_glycan
        self.o_glycan = o_glycan
        self.hypothesis_id = hypothesis_id
        self._source_database = None

    @staticmethod
    def determine_database_glycan_types(path, hypothesis_id=1):
        return _determine_database_contents(path, hypothesis_id=hypothesis_id)

    @property
    def source_database(self):
        if self._source_database is None:
            self._source_database = disk_backed_database.PeptideDiskBackedStructureDatabase(
                self.path, hypothesis_id=self.hypothesis_id)
        return self._source_database

    def session_resolver(self):
        return self.source_database.session

    def hypothesis_resolver(self) -> GlycopeptideHypothesis:
        return self.source_database.hypothesis

    def __reduce__(self):
        return self.__class__, (self.path, self.n_glycan, self.o_glycan, self.hypothesis_id)

    def __call__(self):
        db = disk_backed_database.PeptideDiskBackedStructureDatabase(
            self.path, hypothesis_id=self.hypothesis_id)
        if self.n_glycan and self.o_glycan:
            filter_level = 0
        elif self.n_glycan:
            filter_level = 1
        elif self.o_glycan:
            filter_level = 2
        else:
            raise ValueError("Cannot determine how to filter peptides")
        peptides = []

        self.log("... Loading peptides from %r:%r" % (self.path, self.hypothesis_id))
        start = datetime.datetime.now()
        if filter_level == 1:
            has_sites = db.has_protein_sites()
            # This is an old database, have to do a full scan.
            if not has_sites:
                q = db.having_glycosylation_site()
                iterator = FetchManyIterator(db.session.execute(q))
            else:
                # Fast path for N-glycosylation sites which are marked, but where the bounds
                # aren't precise so the if statements are still needed.
                q = db.spanning_n_glycosylation_site()
                iterator = FetchManyIterator(db.session.execute(q))
        else:
            q = db.having_glycosylation_site()
            iterator = FetchManyIterator(db.session.execute(q))
        seen = set()
        for r in iterator:
            rec = PeptideDatabaseRecord.from_record(r)
            if rec.id in seen:
                self.log("Converted Peptide %d more than once!" % rec.id)
            seen.add(rec.id)
            if filter_level == 1 and rec.n_glycosylation_sites:
                peptides.append(rec)
            elif filter_level == 2 and rec.o_glycosylation_sites:
                peptides.append(rec)
            elif filter_level == 0 and rec.has_glycosylation_sites():
                peptides.append(rec)
        db.session.remove()
        PeptideDatabaseRecord.unshare_sites(peptides)
        end = datetime.datetime.now()
        elapsed = (end - start).total_seconds()
        self.log("... %0.2f seconds elapsed. Loaded %d peptides" % (elapsed, len(peptides)))
        mem_db = disk_backed_database.InMemoryPeptideStructureDatabase(peptides, db)
        return mem_db


def _determine_database_contents(path, hypothesis_id=1):
    db = disk_backed_database.PeptideDiskBackedStructureDatabase(
        path, hypothesis_id=hypothesis_id)
    glycan_classes = db.query(
        serialize.GlycanCombination,
        serialize.GlycanClass.name).join(
            serialize.GlycanCombination.components).join(
            serialize.GlycanComposition.structure_classes).group_by(
            serialize.GlycanClass.name).all()
    glycan_classes = {
        pair[1] for pair in glycan_classes
    }
    n_glycan = serialize.GlycanTypes.n_glycan in glycan_classes
    o_glycan = (serialize.GlycanTypes.o_glycan in glycan_classes or
                serialize.GlycanTypes.gag_linker in glycan_classes)
    return {
        'n_glycan': n_glycan,
        'o_glycan': o_glycan
    }


def make_peptide_glycosylator(path, hypothesis_id: int = 1, target_peptide: bool=True) -> PeptideGlycosylator:
    peptide_db = make_memory_database_proxy_resolver(path, hypothesis_id=hypothesis_id)
    gp_hypothesis: GlycopeptideHypothesis = peptide_db.session.query(
        GlycopeptideHypothesis).get(hypothesis_id)

    glycan_recs = GlycanCombinationRecord.from_hypothesis(
        peptide_db.session, gp_hypothesis.glycan_hypothesis_id)

    generator = PeptideGlycosylator(
        peptide_db,
        glycan_recs,
        default_structure_type=StructureClassification.target_peptide_target_glycan if target_peptide else
                               StructureClassification.decoy_peptide_target_glycan)
    return generator


def make_predictive_glycopeptide_search(path, hypothesis_id: int=1,
                                        target_peptide: bool=True,
                                        product_error_tolerance: float=2e-5,
                                        glycan_score_threshold: float=1.0,
                                        probing_range_for_missing_precursors: int=3,
                                        peptide_masses_per_scan: int=100,
                                        oxonium_ion_threshold: float=0.05) -> PredictiveGlycopeptideSearch:

    generator = make_peptide_glycosylator(
        path, hypothesis_id=hypothesis_id, target_peptide=target_peptide)

    if glycan_score_threshold > 0:
        min_fragments = 2
    else:
        min_fragments = 0

    predictive_search = PredictiveGlycopeptideSearch(
        generator,
        product_error_tolerance=product_error_tolerance,
        glycan_score_threshold=glycan_score_threshold,
        min_fragments=min_fragments,
        probing_range_for_missing_precursors=probing_range_for_missing_precursors,
        trust_precursor_fits=True,
        peptide_masses_per_scan=peptide_masses_per_scan,
        oxonium_ion_threshold=oxonium_ion_threshold)

    return predictive_search


class MultipartGlycopeptideIdentifier(SearchEngineBase):
    precursor_error_tolerance: float
    product_error_tolerance: float
    batch_size: int
    fdr_estimation_strategy: GlycopeptideFDREstimationStrategy
    scorer_type: Type[GlycopeptideSpectrumMatcherBase]
    evaluation_kwargs: Dict[str, Any]
    mass_shifts: List[MassShift]
    probing_range_for_missing_precursors: int
    trust_precursor_fits: bool
    glycan_score_threshold: float
    oxonium_threshold: float
    peptide_masses_per_scan: int

    ipc_manager: SyncManager
    file_manager: TempFileManager
    journal_path: str
    journal_path_collection: List[str]

    target_peptide_db: mass_collection.SearchableMassCollectionWrapper
    decoy_peptide_db: mass_collection.SearchableMassCollectionWrapper
    glycosylation_site_models_path: str

    cache_seeds: Any

    tandem_scans: List[ScanStub]
    scan_loader: Union[ProcessedRandomAccessScanSource,
                       LCMSMSQueryInterfaceMixin]

    n_processes: int

    def __init__(self,
                 tandem_scans,
                 scorer_type,
                 target_peptide_db,
                 decoy_peptide_db,
                 scan_loader,
                 mass_shifts=None,
                 n_processes=6,
                 evaluation_kwargs=None,
                 ipc_manager=None,
                 file_manager=None,
                 probing_range_for_missing_precursors=3,
                 trust_precursor_fits=True,
                 glycan_score_threshold=1.0,
                 peptide_masses_per_scan=60,
                 fdr_estimation_strategy=None,
                 glycosylation_site_models_path=None,
                 cache_seeds=None,
                 oxonium_threshold: float=0.05,
                 **kwargs):
        if fdr_estimation_strategy is None:
            fdr_estimation_strategy = GlycopeptideFDREstimationStrategy.multipart_gamma_gaussian_mixture
        if scorer_type is None:
            scorer_type = LogIntensityScorer
        if evaluation_kwargs is None:
            evaluation_kwargs = {}
        if mass_shifts is None:
            mass_shifts = []
        if Unmodified not in mass_shifts:
            mass_shifts = [Unmodified] + list(mass_shifts)
        if file_manager is None:
            file_manager = TempFileManager()
        elif isinstance(file_manager, str):
            file_manager = TempFileManager(file_manager)
        if ipc_manager is None:
            ipc_manager = make_shared_memory_manager()
        if isinstance(target_peptide_db, str):
            target_peptide_db = self.build_default_disk_backed_db_wrapper(target_peptide_db)
        if isinstance(decoy_peptide_db, str):
            decoy_peptide_db = self.build_default_disk_backed_db_wrapper(decoy_peptide_db)

        self.tandem_scans = tandem_scans
        self.scan_loader = scan_loader

        self.scorer_type = scorer_type
        self.fdr_estimation_strategy = GlycopeptideFDREstimationStrategy[fdr_estimation_strategy]

        self.target_peptide_db = target_peptide_db
        self.decoy_peptide_db = decoy_peptide_db

        self.probing_range_for_missing_precursors = probing_range_for_missing_precursors
        self.trust_precursor_fits = trust_precursor_fits

        self.glycan_score_threshold = glycan_score_threshold
        self.peptide_masses_per_scan = peptide_masses_per_scan

        self.precursor_error_tolerance = 5e-6
        self.product_error_tolerance = 2e-5
        self.batch_size = 1000

        self.mass_shifts = mass_shifts

        self.evaluation_kwargs = evaluation_kwargs
        self.evaluation_kwargs.update(kwargs)

        self.n_processes = n_processes
        self.ipc_manager = ipc_manager
        self.cache_seeds = cache_seeds

        self.file_manager = file_manager
        self.journal_path = self.file_manager.get('glycopeptide-match-journal')
        self.journal_path_collection = []
        self.glycosylation_site_models_path = glycosylation_site_models_path

        self.oxonium_threshold = oxonium_threshold

    @classmethod
    def build_default_disk_backed_db_wrapper(cls, path, **kwargs):
        peptide_db = make_disk_backed_peptide_database(path, **kwargs)
        return peptide_db

    @classmethod
    def build_default_memory_backed_db_wrapper(cls, path, **kwargs):
        peptide_db = make_memory_database_proxy_resolver(path, **kwargs)
        return peptide_db

    def build_scan_groups(self):
        if self.tandem_scans is None:
            pinfos = self.scan_loader.precursor_information()
            self.tandem_scans = [ScanStub(pinfo, self.scan_loader) for pinfo in pinfos]
        groups = group_by_precursor_mass(self.tandem_scans, 1e-4)
        return groups

    def build_predictive_searchers(self):
        glycan_combinations = GlycanCombinationRecord.from_hypothesis(
            self.target_peptide_db.session, self.target_peptide_db.hypothesis_id)

        if self.glycan_score_threshold > 0:
            min_fragments = 2
        else:
            min_fragments = 0

        glycan_prior_model = None
        if self.glycosylation_site_models_path is not None:
            self.log("Loading glycosylation site scoring models from %r" % self.glycosylation_site_models_path)
            glycan_prior_model = GlycoproteomePriorAnnotator.load(
                self.target_peptide_db.session,
                self.decoy_peptide_db.session,
                open(self.glycosylation_site_models_path, 'rt'))

        expand_combinatorics = True

        generator = PeptideGlycosylator(
            self.target_peptide_db,
            glycan_combinations,
            default_structure_type=StructureClassification.target_peptide_target_glycan,
            glycan_prior_model=glycan_prior_model,
            expand_combinatorics=expand_combinatorics
        )
        target_predictive_search = PredictiveGlycopeptideSearch(
            generator,
            product_error_tolerance=self.product_error_tolerance,
            glycan_score_threshold=self.glycan_score_threshold,
            min_fragments=min_fragments,
            probing_range_for_missing_precursors=self.probing_range_for_missing_precursors,
            trust_precursor_fits=self.trust_precursor_fits,
            peptide_masses_per_scan=self.peptide_masses_per_scan,
            oxonium_ion_threshold=self.oxonium_threshold)

        generator = PeptideGlycosylator(
            self.decoy_peptide_db,
            glycan_combinations,
            default_structure_type=StructureClassification.decoy_peptide_target_glycan,
            glycan_prior_model=glycan_prior_model,
            expand_combinatorics=expand_combinatorics
        )
        decoy_predictive_search = PredictiveGlycopeptideSearch(
            generator,
            product_error_tolerance=self.product_error_tolerance,
            glycan_score_threshold=self.glycan_score_threshold,
            min_fragments=min_fragments,
            probing_range_for_missing_precursors=self.probing_range_for_missing_precursors,
            trust_precursor_fits=self.trust_precursor_fits,
            peptide_masses_per_scan=self.peptide_masses_per_scan,
            oxonium_ion_threshold=self.oxonium_threshold)
        return target_predictive_search, decoy_predictive_search

    def run_branching_identification_pipeline(self, scan_groups):
        (target_predictive_search,
         decoy_predictive_search) = self.build_predictive_searchers()

        spectrum_batcher = SpectrumBatcher(
            scan_groups,
            Queue(10),
            max_scans_per_workload=self.batch_size)

        spectrum_batcher.start(daemon=True)

        common_queue = multiprocessing.Queue(5)
        label_to_batch_queue = {
            "target": common_queue,
            "decoy": common_queue,
        }

        ipc_logger = IPCLoggingManager()

        mapping_batcher = BatchMapper(
            # map labels to be loaded in the mapper executor to avoid repeatedly
            # serializing the databases.
            [
                ('target', 'target'),
                ('decoy', 'decoy')
            ],
            spectrum_batcher.out_queue,
            label_to_batch_queue,
            spectrum_batcher.done_event,
            precursor_error_tolerance=self.precursor_error_tolerance,
            mass_shifts=self.mass_shifts)
        mapping_batcher.done_event = multiprocessing.Event()
        mapping_batcher.start(daemon=True)

        execution_branches: List['IdentificationWorker'] = []
        scorer_type_payload = zlib.compress(pickle.dumps(self.scorer_type, -1), 9)
        predictive_search_payload = zlib.compress(
            pickle.dumps(
                (target_predictive_search, decoy_predictive_search), -1), 9)

        for i in range(self.n_processes):
            journal_path_for = self.file_manager.get('glycopeptide-match-journal-%d' % (i))
            self.journal_path_collection.append(journal_path_for)
            done_event_for = multiprocessing.Event()
            branch = IdentificationWorker(
                "IdentificationWorker-%d" % i,
                self.ipc_manager.address,
                common_queue,
                mapping_batcher.done_event,
                journal_path_for,
                done_event_for,
                self.scan_loader,
                target_predictive_search=predictive_search_payload,
                decoy_predictive_search=None,
                scorer_type=scorer_type_payload,
                evaluation_kwargs=self.evaluation_kwargs,
                error_tolerance=self.product_error_tolerance,
                cache_seeds=self.cache_seeds,
                mass_shifts=self.mass_shifts,
                log_handler=ipc_logger.sender(),
            )
            execution_branches.append(branch)
        # Clear these big blobs from the parent process, we no longer need them
        del scorer_type_payload
        del predictive_search_payload
        pipeline = Pipeline([
            spectrum_batcher,
            mapping_batcher,
        ] + execution_branches)

        for branch in execution_branches:
            try:
                branch.start(process=True, daemon=True)
            except Exception as err:
                self.log(f"... Error {err} occurred during worker startup")

        # At this point, all the components have started already, but
        # to let the Pipeline object setup its "started" invariants, call
        # `start` again, which is a no-op for already-started task sequences.
        pipeline.start(daemon=True)
        pipeline.join()
        ipc_logger.stop()
        had_error = pipeline.error_occurred()
        if had_error:
            message = "%d unrecoverable error%s occured during search!" % (
                had_error, 's' if had_error > 1 else '')
            pipeline.stop()
            common_queue.close()
            common_queue.cancel_join_thread()
            self.log("Terminating search pipeline")
            for branch in execution_branches:
                branch.kill_process()
            raise Exception(message)
        total = 0
        for branch in execution_branches:
            total += branch.results_processed.value
        return total

    def _load_identifications_from_journal(
            self, journal_path: os.PathLike,
            total_solutions_count: int,
            accumulator: Optional[List[MultiScoreSpectrumMatch]]=None) -> List[MultiScoreSpectrumMatch]:
        if accumulator is None:
            accumulator = []

        with self.scan_loader.toggle_peak_loading():
            reader = enumerate(JournalFileReader(
                journal_path,
                scan_loader=ScanInformationLoader(self.scan_loader),
                mass_shift_map={m.name: m for m in self.mass_shifts},
                score_set_type=self.scorer_type.get_score_set_type()),
                len(accumulator))
            i = float(len(accumulator))
            try:
                # Get the nearest progress checkpoint
                last = round(i / total_solutions_count, 1)
            except ZeroDivisionError:
                last = 0.1
            should_log = False
            for i, sol in reader:
                if i * 1.0 / total_solutions_count > last:
                    should_log = True
                    last += 0.1
                elif i % 100000 == 0 and i > 1:
                    should_log = True
                if should_log:
                    self.log("... %d/%d Solutions Loaded (%0.2f%%)" % (
                        i, total_solutions_count, i * 100.0 / total_solutions_count))
                    should_log = False
                accumulator.append(sol)
        return accumulator

    def search(self, precursor_error_tolerance: float=1e-5,
               simplify: bool=True,
               batch_size: int=500, **kwargs) -> SolutionSetGrouper:
        self.evaluation_kwargs.update(kwargs)
        self.product_error_tolerance = self.evaluation_kwargs.pop('error_tolerance', 2e-5)
        self.precursor_error_tolerance = precursor_error_tolerance
        self.batch_size = batch_size

        self.log("Building Scan Groups...")
        scan_groups = self.build_scan_groups()
        self.log("{:d} Scans, {:d} Scan Groups".format(
            len(self.tandem_scans), len(scan_groups)))
        self.log("Running Identification Pipeline...")
        start_time = datetime.datetime.now()
        total_solutions_count = self.run_branching_identification_pipeline(scan_groups)
        end_time = datetime.datetime.now()
        self.log("Database Search Complete, %s Elapsed" % (end_time - start_time))
        self.log("Loading Spectrum Matches From Journal...")
        solutions = []
        n = len(self.journal_path_collection)
        for i, journal_path in enumerate(self.journal_path_collection, 1):
            self.log("... Reading Journal Shard %s, %d/%d" % (journal_path, i, n))
            self._load_identifications_from_journal(journal_path, total_solutions_count, solutions)
        self.log("Partitioning Spectrum Matches...")
        groups = SolutionSetGrouper(solutions)
        return groups

    def estimate_fdr(self, glycopeptide_spectrum_match_groups: SolutionSetGrouper,
                     *args, **kwargs) -> Tuple[SolutionSetGrouper, GlycopeptideFDREstimator]:
        keys = [SpectrumMatchClassification[i] for i in range(4)]
        g = glycopeptide_spectrum_match_groups
        self.log("Running Target Decoy Analysis with %d targets and %d/%d/%d decoys" % (
            len(g[keys[0]]), len(g[keys[1]]), len(g[keys[2]]), len(g[keys[3]])))
        peptide_fdr_estimator = self.scorer_type.get_fdr_model_for_dimension('peptide')
        estimator = GlycopeptideFDREstimator(
            glycopeptide_spectrum_match_groups,
            self.fdr_estimation_strategy,
            peptide_fdr_estimator=peptide_fdr_estimator)
        groups: SolutionSetGrouper = estimator.start()
        self.log("Rebuilding Targets")
        cache = {}
        target_match_sets = groups.target_matches
        n = len(target_match_sets)
        for i, target_match_set in enumerate(target_match_sets):
            if i % 10000 == 0 and i:
                self.log("... Rebuilt %d Targets (%0.2f%%)" % (i, i * 100.0 / n))
            for target_match in target_match_set:
                if target_match.target.id in cache:
                    target_match.target = cache[target_match.target.id]
                else:
                    target_match.target = cache[target_match.target.id] = target_match.target.convert()
        cache.clear()
        return groups, estimator

    def map_to_chromatograms(
            self,
            chromatograms: List[Chromatogram],
            tandem_identifications: List[MultiScoreSpectrumSolutionSet],
            precursor_error_tolerance: float=1e-5,
            threshold_fn: Callable[[SpectrumMatch], bool]=lambda x: x.q_value < 0.05,
            entity_chromatogram_type: Type[GlycopeptideChromatogram]=GlycopeptideChromatogram
            ) -> Tuple[List[TandemAnnotatedChromatogram], List[TandemSolutionsWithoutChromatogram]]:
        self.log("Mapping MS/MS Identifications onto Chromatograms")
        self.log("%d Chromatograms" % len(chromatograms))
        mapper = ChromatogramMSMSMapper(
            chromatograms, precursor_error_tolerance,
            self.scan_loader.convert_scan_id_to_retention_time)
        self.log("Assigning Solutions")
        mapper.assign_solutions_to_chromatograms(tandem_identifications)
        self.log("Distributing Orphan Spectrum Matches")
        mapper.distribute_orphans(threshold_fn=threshold_fn)
        self.log("Selecting Most Representative Matches")
        mapper.assign_entities(threshold_fn, entity_chromatogram_type=entity_chromatogram_type)
        return mapper.chromatograms, mapper.orphans


class IdentificationWorker(TaskExecutionSequence):
    name: str
    ipc_manager_address: Union[str, Tuple[str, int]]

    input_batch_queue: multiprocessing.Queue
    input_done_event: multiprocessing.Event

    done_event: multiprocessing.Event

    journal_path: os.PathLike
    scan_loader: Union[ProcessedRandomAccessScanSource,
                       LCMSMSQueryInterfaceMixin]

    target_predictive_search: PredictiveGlycopeptideSearch
    decoy_predictive_search: PredictiveGlycopeptideSearch

    n_processes: int
    scorer_type: Type[GlycopeptideSpectrumMatcherBase]

    scorer_type: Type[GlycopeptideSpectrumMatcherBase]
    evaluation_kwargs: Dict[str, Any]
    error_tolerance: float
    cache_seeds: Any
    mass_shifts: List[MassShift]

    log_handler: LoggingHandlerToken

    def __init__(self, name, ipc_manager_address, input_batch_queue, input_done_event,
                 journal_path, done_event,
                 # Mapping Executor Parameters
                 scan_loader=None, target_predictive_search=None, decoy_predictive_search=None,
                 # Matching Executor Parameters
                 n_processes=1, scorer_type=None, evaluation_kwargs=None, error_tolerance=None,
                 cache_seeds=None, mass_shifts=None,
                 log_handler=None):
        self.name = name
        self.ipc_manager_address = ipc_manager_address
        self.input_batch_queue = input_batch_queue
        self.input_done_event = input_done_event
        self.journal_path = journal_path
        self.done_event = done_event

        # Mapping Executor
        self.scan_loader = scan_loader
        self.target_predictive_search = target_predictive_search
        self.decoy_predictive_search = decoy_predictive_search

        # Matching Executor
        self.n_processes = n_processes
        self.scorer_type = scorer_type
        self.evaluation_kwargs = evaluation_kwargs
        self.error_tolerance = error_tolerance
        self.cache_seeds = cache_seeds
        self.mass_shifts = mass_shifts
        self.results_processed = multiprocessing.Value(ctypes.c_uint64)
        self.log_handler = log_handler

    def _get_repr_details(self):
        props = ["name=%r" % self.name, "pid=%r" % multiprocessing.current_process().pid]
        return ', '.join(props)

    def _name_for_execution_sequence(self):
        return self.name

    def run(self):
        self.try_set_process_name("glycresoft-identification")
        self.log_handler.add_handler()
        ipc_manager = SyncManager(self.ipc_manager_address)
        ipc_manager.connect()
        lock = threading.RLock()
        # Late loading of the compressed serialized scorer type to avoid balooning
        # the parent process
        if isinstance(self.scorer_type, (str, bytes)):
            self.scorer_type = pickle.loads(zlib.decompress(self.scorer_type))
        if isinstance(self.target_predictive_search, (str, bytes)):
            self.target_predictive_search, self.decoy_predictive_search = pickle.loads(
                zlib.decompress(self.target_predictive_search))

        oxonium_ion_index = OxoniumIndex()
        oxonium_ion_index.build_index(
            self.target_predictive_search.peptide_glycosylator.glycan_combinations, oxonium=True)

        signatures = list(single_signatures)
        if self.evaluation_kwargs.get("rare_signatures"):
            signatures.extend(compound_signatures)
        signature_index = SignatureIonIndex(signatures)
        signature_index.build_index(
            self.target_predictive_search.peptide_glycosylator.glycan_combinations)

        self.target_predictive_search.oxonium_ion_index = oxonium_ion_index
        self.target_predictive_search.signature_ion_index = signature_index
        self.decoy_predictive_search.oxonium_ion_index = oxonium_ion_index
        self.decoy_predictive_search.signature_ion_index = signature_index

        mapping_executor = SemaphoreBoundMapperExecutor(
            lock,
            OrderedDict([
                ('target', self.target_predictive_search),
                ('decoy', self.decoy_predictive_search)
            ]),
            self.scan_loader,
            self.input_batch_queue,
            Queue(5),
            self.input_done_event,
        )
        matching_executor = SemaphoreBoundMatcherExecutor(
            lock,
            mapping_executor.out_queue,
            Queue(5),
            mapping_executor.done_event,
            scorer_type=self.scorer_type,
            ipc_manager=ipc_manager,
            n_processes=self.n_processes,
            mass_shifts=self.mass_shifts,
            evaluation_kwargs=self.evaluation_kwargs,
            error_tolerance=self.error_tolerance,
            cache_seeds=self.cache_seeds
        )

        journal_writer = JournalFileWriter(
            self.journal_path, score_columns=self.scorer_type.get_score_set_type().field_names())
        journal_consumer = JournalingConsumer(
            journal_writer,
            matching_executor.out_queue,
            matching_executor.done_event)
        journal_consumer.done_event = self.done_event

        pipeline = Pipeline([
            mapping_executor,
            matching_executor,
            journal_consumer,
        ])
        pipeline.start(daemon=True)
        pipeline.join()
        if pipeline.error_occurred():
            self.log("An error occurred while executing %s" % (self, ))
            self.set_error_occurred()
        journal_writer.close()
        self.results_processed.value = journal_writer.solution_counter
        self.log("%s has finished. %d solutions calculated." %
                 (self, journal_writer.solution_counter, ))



import itertools

from collections import deque, defaultdict
from typing import List, Set, Optional, Dict, DefaultDict, Type, ClassVar, Generic, TypeVar, Deque, Any

import numpy as np

from glypy.structure.glycan_composition import FrozenMonosaccharideResidue
from ms_deisotope.peak_dependency_network.intervals import SpanningMixin

from .grouping import ChromatogramRetentionTimeInterval, IntervalTreeNode
from .index import ChromatogramFilter
from .chromatogram import Chromatogram


_standard_transitions = [
    FrozenMonosaccharideResidue.from_iupac_lite("HexNAc"),
    FrozenMonosaccharideResidue.from_iupac_lite("Hex"),
    FrozenMonosaccharideResidue.from_iupac_lite('NeuAc'),
    FrozenMonosaccharideResidue.from_iupac_lite("Fuc"),
    FrozenMonosaccharideResidue.from_iupac_lite("HexA"),
]


class UnknownTransition(object):
    def __init__(self, mass):
        self._mass = mass
        self._hash = hash(mass)

    def __eq__(self, other):
        return self.mass() == other.mass()

    def mass(self):
        return self._mass

    def __hash__(self):
        return self._hash

    def __repr__(self):
        return "{self.__class__.__name__}({mass})".format(self=self, mass=self.mass())


class TimeQuery(SpanningMixin):
    def __init__(self, chromatogram, width=0):
        self.start = chromatogram.start_time - width
        self.end = chromatogram.end_time + width

    def __repr__(self):
        return "TimeQuery(%f, %f)" % (self.start, self.end)


class ChromatogramGraphNode(ChromatogramRetentionTimeInterval):
    chromatogram: Chromatogram
    edges: Set['ChromatogramGraphEdge']
    index: int
    center: float

    def __init__(self, chromatogram, index, edges=None):
        super(ChromatogramGraphNode, self).__init__(chromatogram)
        if edges is None:
            edges = set()
        self.chromatogram = chromatogram
        self.index = index
        self.edges = edges

        total = 0
        abundance = 0
        for node in self.chromatogram.nodes:
            intensity = node.total_intensity()
            total += node.retention_time * intensity
            abundance += intensity
        self.center = total / abundance

    def get_chromatogram(self):
        return self.chromatogram

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.chromatogram,)

    def __index__(self):
        return self.index

    def __hash__(self):
        return hash(self.index)

    def __eq__(self, other):
        return self.chromatogram == other.chromatogram


class ChromatogramGraphEdge(object):
    node_a: ChromatogramGraphNode
    node_b: ChromatogramGraphNode
    weight: float
    mass_error: float
    rt_error: float

    def __init__(self, node_a, node_b, transition, weight=1, mass_error=0, rt_error=0):
        self.node_a = node_a
        self.node_b = node_b
        self.transition = transition
        self.weight = weight
        self.mass_error = mass_error
        self.rt_error = rt_error
        self.node_a.edges.add(self)
        self.node_b.edges.add(self)

    def __repr__(self):
        return "%s(%s, %s, %s)" % (
            self.__class__.__name__, self.node_a.chromatogram, self.node_b.chromatogram,
            self.transition)

    def __hash__(self):
        return hash(frozenset((self.node_a.index, self.node_b.index)))

    def __eq__(self, other):
        return frozenset((self.node_a.index, self.node_b.index)) == frozenset((other.node_a.index, other.node_b.index))

    def __ne__(self, other):
        return not self == other


def explode_node(node):
    items = deque()
    results = []

    for edge in node.edges:
        items.append(edge.node_a)
        items.append(edge.node_b)

    visited = set()
    while items:
        node = items.popleft()
        if node.index in visited:
            continue
        visited.add(node.index)
        results.append(node)

        for edge in node.edges:
            if edge.node_a.index not in visited:
                items.append(edge.node_a)
            if edge.node_b.index not in visited:
                items.append(edge.node_b)
    return results


def chromatograms_from_edge(edge):
    return edge.node_a.chromatogram, edge.node_b.chromatogram


V = TypeVar("V", bound=ChromatogramGraphNode)
E = TypeVar("E", bound=ChromatogramGraphEdge)

class ChromatogramGraph(Generic[V, E]):
    node_cls: ClassVar[Type[V]] = ChromatogramGraphNode
    edge_cls: ClassVar[Type[E]] = ChromatogramGraphEdge

    chromatograms: List[Chromatogram]
    assigned_seed_queue: Deque[V]
    assignment_map: Dict[Any, V]

    nodes: List[V]
    edges: Set[E]

    rt_tree: IntervalTreeNode

    def __init__(self, chromatograms):
        self.chromatograms = chromatograms
        self.assigned_seed_queue = deque()
        self.assignment_map = {}
        self.nodes = self._construct_graph_nodes(self.chromatograms)
        self.rt_tree = IntervalTreeNode.build(self.nodes)
        self.edges = set()

    def __len__(self):
        return len(self.nodes)

    def __iter__(self):
        return iter(self.nodes)

    def __getitem__(self, i):
        return self.nodes[i]

    def _construct_graph_nodes(self, chromatograms) -> List[V]:
        nodes = []
        for i, chroma in enumerate(chromatograms):
            node = (self.node_cls(chroma, i))
            nodes.append(node)
            if node.chromatogram.composition:
                self.enqueue_seed(node)
                self.assignment_map[node.chromatogram.composition] = node
        return nodes

    def enqueue_seed(self, chromatogram_with_assignment):
        self.assigned_seed_queue.append(chromatogram_with_assignment)

    def pop_seed(self):
        return self.assigned_seed_queue.popleft()

    def iterseeds(self):
        while self.assigned_seed_queue:
            yield self.pop_seed()

    def find_edges(self, node: V, query_width: float=2., transitions=None, **kwargs):
        if transitions is None:
            transitions = _standard_transitions
        query = TimeQuery(node.chromatogram, query_width)
        nodes = ChromatogramFilter(self.rt_tree.overlaps(query.start, query.end))

        starting_mass = node.neutral_mass

        for transition in transitions:
            added = starting_mass + transition.mass()
            match = nodes.find_mass(added)
            if match:
                ppm_error = (added - match.neutral_mass) / match.neutral_mass
                rt_error = (node.center - match.center)
                self.edges.add(self.edge_cls(node, match, transition, mass_error=ppm_error, rt_error=rt_error))

    def find_shared_peaks(self):
        peak_map = defaultdict(set)
        for chromatogram_node in self.nodes:
            for peak_bunch in chromatogram_node.chromatogram.peaks:
                for peak in peak_bunch:
                    peak_map[peak].add(chromatogram_node)
        edges = set()
        for peak, nodes in peak_map.items():
            for a, b in itertools.combinations(nodes, 2):
                edges.add(frozenset((a, b)))

        result = []
        for edge in edges:
            start, end = sorted(edge, key=lambda x: x.index)
            delta = -(start.neutral_mass - end.neutral_mass)
            chromatogram_edge = ChromatogramGraphEdge(
                start, end, UnknownTransition(delta), rt_error=start.center - end.center)
            result.append(chromatogram_edge)
            self.edges.add(chromatogram_edge)
        return result

    def build(self, query_width: float=2., transitions=None, **kwargs):
        for node in self.iterseeds():
            self.find_edges(node, query_width=query_width, transitions=transitions, **kwargs)

    def adjacency_matrix(self) -> np.ndarray:
        adjmat = np.zeros((len(self), len(self)))
        nodes = self.nodes
        for node in nodes:
            for edge in node.edges:
                adjmat[edge.node_a.index, edge.node_b.index] = 1
        return adjmat

    def connected_components(self) -> List[List[V]]:
        pool = set(self.nodes)
        components = []

        i = 0
        while pool:
            i += 1
            current_component = set()
            visited = set()
            node = pool.pop()
            current_component.add(node)
            j = 0
            while current_component:
                j += 1
                node = current_component.pop()
                visited.add(node)
                for edge in node.edges:
                    if edge.node_a not in visited and edge.node_a in pool:
                        current_component.add(edge.node_a)
                        pool.remove(edge.node_a)
                    if edge.node_b not in visited and edge.node_b in pool:
                        current_component.add(edge.node_b)
                        pool.remove(edge.node_b)
            components.append(list(visited))
        return components



import logging

import numpy as np

from glypy import GlycanComposition

from glycresoft.database.composition_network import (
    CompositionGraphNode, NeighborhoodWalker)

from glycresoft.task import log_handle

from .constants import (
    DEFAULT_LAPLACIAN_REGULARIZATION,
    NORMALIZATION,
    DEFAULT_RHO,
    RESET_THRESHOLD_VALUE)

from .laplacian_smoothing import LaplacianSmoothingModel, ProportionMatrixNormalization

from .graph import (
    BlockLaplacian,
    assign_network,
    network_indices,
    weighted_laplacian_matrix)

from .grid_search import (
    NetworkReduction,
    NetworkTrimmingSearchSolution,
    ThresholdSelectionGridSearch)

from .observation import (
    GlycanCompositionSolutionRecord,
    VariableObservationAggregation)

from .utils import fast_positive_definite_inverse

logger = logging.getLogger("glycresoft.glycome_network_smoothing")
logger.addHandler(logging.NullHandler())


def _has_glycan_composition(x):
    try:
        gc = x.glycan_composition
        return gc is not None
    except AttributeError:
        return False


class GlycomeModel(LaplacianSmoothingModel):
    """An implementation of the Glycan Network Smoothing by Laplacian Regularization.

    Attributes
    ----------
    block_L: :class:`~.BlockLaplacian`
        The block-oriented version of the Weighted Laplacian matrix of :attr:`network`
    network: :class:`~.CompositionGraph`
        A graph of glycan compositions that has been annotated by their observation confidence.
        This network may be pruned during the course of model fitting. The original network is
        always available under :attr:`_network`.
    observed_compositions: list of :class:`~.GlycanCompositionSolutionRecord`
        The observed and scored glycan compositions. These may not be unique, and will be
        summarized prior to being projected onto the network by :attr:`observation_aggregator`.
        This list may be truncated during the course of model fitting. The original list is always
        available under :attr:`_observed_compositions`.
    threshold: float
        The current threshold applied to the observed score. This value may change during model
        fitting.
    belongingness_matrix: np.ndarray[float, ndim=2]
        A C by N matrix where C is the number of glycan compositions in the network and N is the
        number of neighborhoods over the network, with the value at [i, j] corresponding to how much
        C_i belongs to N_j. This value is not normalized, see :attr:`normalized_belongingness_matrix`.
    normalized_belongingness_matrix: np.ndarray[float, ndim=2]
        A normalized version of :attr:`belongingness_matrix`, which follows the normalization paradigm
        given by :attr:`_belongingness_normalization`
    A0: :class:`np.ndarray[float, ndim=2]`
        The subset of :attr:`normalized_belongingness_matrix` corresponding to :attr:`observed_compositions`
    Am: :class:`np.ndarray[float, ndim=2]`
        The subset of :attr:`normalized_belongingness_matrix` corresponding to nodes in :attr:`network` that
        are not in :attr:`observed_compositions`
    S0:  :class:`np.ndarray[float]`
        The summarized scores of the observed nodes in :attr:`network`
    C0: list of :class:`CompositionGraphNode`
        The observed nodes in :attr:`network`
    obs_ix: :class:`np.ndarray[int]`
        The indices into :attr:`network` that were observed
    miss_ix: :class:`np.ndarray[int]`
        The indices into :attr:`network` that were not observed
    summarized_state: :class:`~.ObservationWeightState`
        A helper intermediary which holds the transformation from :attr:`observed_compositions`
        to :attr:`S0`, :attr:`variance_matrix` and :attr:`inverse_variance_matrix`
    variance_matrix: :class:`np.ndarray[float, ndim=2]`
        The variance of the observed scores, as estimated by :attr:`observation_aggregator`
        from :attr:`observed_compositions`.
    inverse_variance_matrix: :class:`np.ndarray[float, ndim=2]`
        The inverse of :attr:`variance_matrix`, computed separately for efficiency.
    """
    def __init__(self, observed_compositions, network, belongingness_matrix=None,
                 regularize=DEFAULT_LAPLACIAN_REGULARIZATION,
                 belongingness_normalization=NORMALIZATION,
                 observation_aggregator=VariableObservationAggregation):
        self.observation_aggregator = observation_aggregator
        observed_compositions = [
            o for o in observed_compositions if _has_glycan_composition(o) and o.score > 0]
        self._observed_compositions = observed_compositions
        self._configure_with_network(network)
        if len(self.miss_ix) == 0:
            self._network.add_node(CompositionGraphNode(GlycanComposition(), -1), reindex=True)
            self._configure_with_network(self._network)

        self.block_L = BlockLaplacian(self.network, regularize=regularize)
        self.threshold = self.block_L.threshold

        # Initialize Names
        self.normalized_belongingness_matrix = None
        self.A0 = None
        self._belongingness_normalization = None
        self.S0 = np.array([])
        self.C0 = []
        self.variance_matrix = None
        self.inverse_variance_matrix = None

        # Expensive Step
        if belongingness_matrix is None:
            self.belongingness_matrix = self.build_belongingness_matrix()
        else:
            self.belongingness_matrix = np.array(belongingness_matrix)

        # Normalize and populate
        self.normalize_belongingness(belongingness_normalization)
        self._populate(self._observed_compositions)

    def _configure_with_network(self, network):
        self._network = network
        self.network = assign_network(network.clone(), self._observed_compositions)

        self.neighborhood_walker = NeighborhoodWalker(self.network)

        self.neighborhood_names = self.neighborhood_walker.neighborhood_names()
        self.node_names = [str(node) for node in self._network]

        self.obs_ix, self.miss_ix = network_indices(self.network)

    def __reduce__(self):
        return self.__class__, (
            self._observed_compositions, self._network, self.belongingness_matrix,
            self.block_L.regularize, self._belongingness_normalization,
            self.observation_aggregator)

    def _populate(self, observations):
        var_agg = self.observation_aggregator(self._network)
        var_agg.collect(observations)
        aggregated_observations, summarized_state = var_agg.build_records()
        self.network = assign_network(self._network.clone(), aggregated_observations)
        self.obs_ix, self.miss_ix = network_indices(self.network)

        self.A0 = self.normalized_belongingness_matrix[self.obs_ix, :]
        self.Am = self.normalized_belongingness_matrix[self.miss_ix, :]
        self.S0 = np.array([g.score for g in self.network[self.obs_ix]])
        self.C0 = ([g for g in self.network[self.obs_ix]])
        self.summarized_state = summarized_state
        self.variance_matrix = np.diag(summarized_state.variance_matrix[self.obs_ix, self.obs_ix])
        self.inverse_variance_matrix = np.diag(summarized_state.inverse_variance_matrix[self.obs_ix, self.obs_ix])

    def set_threshold(self, threshold):
        accepted = [
            g for g in self._observed_compositions if g.score > threshold]
        if len(accepted) == 0:
            raise ValueError("Threshold %f produces an empty observed set" % (threshold,))
        self._populate(accepted)
        self.block_L = BlockLaplacian(self.network, threshold=threshold, regularize=self.block_L.regularize)
        self.threshold = self.block_L.threshold

    def reset(self):
        self.set_threshold(RESET_THRESHOLD_VALUE)

    def normalize_belongingness(self, method=NORMALIZATION):
        self.normalized_belongingness_matrix = ProportionMatrixNormalization.normalize(
            self.belongingness_matrix, method)
        self._belongingness_normalization = method
        self.A0 = self.normalized_belongingness_matrix[self.obs_ix, :]

    def apply_belongingness_patch(self):
        updated_belongingness = self.get_belongingness_patch()
        self.normalized_belongingness_matrix = updated_belongingness
        self.A0 = self.normalized_belongingness_matrix[self.obs_ix, :]

    def remove_belongingness_patch(self):
        self.normalized_belongingness_matrix = ProportionMatrixNormalization.normalize(
            self.belongingness_matrix, self._belongingness_normalization)
        self.A0 = self.normalized_belongingness_matrix[self.obs_ix, :]

    def sample_tau(self, rho, lmda):
        sigma_est = np.std(self.S0)
        mu_tau = self.estimate_tau_from_S0(rho, lmda)
        return np.random.multivariate_normal(mu_tau, np.eye(len(mu_tau)).dot(sigma_est ** 2))

    def sample_phi_given_tau(self, tau, lmda):
        return np.random.multivariate_normal(self.A0.dot(tau), (1. / lmda) * self.L_oo_inv)

    def find_optimal_lambda(self, rho, lambda_max=1, step=0.01, threshold=0.0001, fit_tau=True,
                            drop_missing=True, renormalize_belongingness=NORMALIZATION):
        obs = []
        missed = []
        network = self.network.clone()
        for node in network:
            if node.score < threshold:
                missed.append(node)
                node.marked = True
            else:
                obs.append(node.score)
        lambda_values = np.arange(0.01, lambda_max, step)
        press = []
        if drop_missing:
            for node in missed:
                network.remove_node(node, limit=5)
        # The network passed into LaplacianSmoothingModel will have its indices changed,
        # and will not match the ordering of the belongingness matrix, so make sure the
        # observed indices are aligned.
        obs_ix, _miss_ix = network_indices(network)
        wpl = weighted_laplacian_matrix(network)
        lum = LaplacianSmoothingModel(
            network, self.normalized_belongingness_matrix[obs_ix, :], threshold,
            neighborhood_walker=self.neighborhood_walker,
            belongingness_normalization=renormalize_belongingness,
            variance_matrix=self.variance_matrix)
        ident = np.eye(wpl.shape[0])
        for lambd in lambda_values:
            if fit_tau:
                tau = lum.estimate_tau_from_S0(rho, lambd)
            else:
                tau = np.zeros(self.A0.shape[1])
            T = lum.optimize_observed_scores(lambd, lum.A0.dot(tau))
            A = ident + lambd * wpl
            # H = np.linalg.inv(A)
            H = fast_positive_definite_inverse(A)
            press_value = sum(
                ((obs - T) / (1 - (np.diag(H) - np.finfo(float).eps))) ** 2) / len(obs)
            press.append(press_value)
        return lambda_values, np.array(press)

    def find_threshold_and_lambda(self, rho, lambda_max=1., lambda_step=0.02, threshold_start=0.,
                                  threshold_step=0.2, fit_tau=True, drop_missing=True,
                                  renormalize_belongingness=NORMALIZATION):
        r'''Iterate over score thresholds and smoothing factors (lambda), sampling points
        from the parameter grid and computing the PRESS residual at each point.

        This produces a :class:`NetworkReduction` data structure recording the results for
        later local maximum detection.

        Parameters
        ----------
        rho: float
            The scale of the variance of the observed score
        lambda_max: float
            The maximum value of lambda to consider on the grid
        lambda_step: float
            The size of the change in lambda at each iteration
        threshold_start: float
            The minimum observed score threshold to start the grid search at
        threshold_step: float
            The size of the change in the observed score threshold at each iteration
        fit_tau: bool
            Whether or not to estimate :math:`\tau` for each iteration when computing
            the PRESS
        drop_missing: bool
            Whether or not to remove nodes from the graph which are not observed above
            the threshold, restructuring the graph, which in turn changes the Laplacian.
        renormalize_belongingness: str
            A string constant which names the belongingness normalization technique to
            use.

        Returns
        -------
        :class:`NetworkReduction`:
            The recorded grid of sampled points and snapshots of the model at each point
        '''
        solutions = NetworkReduction()
        limit = max(self.S0)
        start = max(min(self.S0) - 1e-3, threshold_start)
        current_network = self.network.clone()
        thresholds = np.arange(start, limit, threshold_step)
        last_solution = None
        last_raw_observations = None
        last_aggregate = None
        for i_threshold, threshold in enumerate(thresholds):
            if i_threshold % 10 == 0:
                logger.info("... Threshold = %r (%0.2f%%)" % (
                    threshold, (100.0 * i_threshold / len(thresholds))))
            # Aggregate the raw observations into averaged, variance reduced records
            # and annotate the network with these new scores
            raw_observations = [c for c in self._observed_compositions if c.score > threshold]

            # cache on the explicit raw observations used because the step size may be smaller than
            # the next highest difference, and aggregating observations can be expensive. There is
            # no solution to the general problem as it calls for inverting a potentially large matrix
            # to only be used in this loop.
            if raw_observations == last_raw_observations:
                observations, summarized_state, obs_ix = last_aggregate # pylint: disable=unpacking-non-sequence
            else:
                agg = self.observation_aggregator(self.network)
                agg.collect(raw_observations)

                observations, summarized_state = agg.build_records()
                obs_ix = agg.observed_indices()
                last_aggregate = (observations, summarized_state, obs_ix)
                last_raw_observations = raw_observations

            # Extract pre-calculated variance matrices
            variance_matrix = summarized_state.variance_matrix
            inverse_variance_matrix = summarized_state.inverse_variance_matrix
            variance_matrix = np.diag(variance_matrix[obs_ix, obs_ix])
            inverse_variance_matrix = np.diag(inverse_variance_matrix[obs_ix, obs_ix])

            # clear the scores from the network
            current_network = current_network.clone()
            for node in current_network:
                node.score = 0
                node.internal_score = 0

            # assign aggregated scores to the network
            network = assign_network(current_network, observations)

            # Filter the network, marking nodes for removal and recording observed
            # nodes for future use.
            obs = []
            missed = []
            for i, node in enumerate(network):
                if node.score < threshold:
                    missed.append(node)
                    node.marked = True
                else:
                    obs.append(node.score)
            if len(obs) == 0:
                break
            obs = np.array(obs)
            press = []

            if drop_missing:
                # drop nodes whose score does not exceed the threshold
                for node in missed:
                    network.remove_node(node, limit=5)

            if last_solution is not None:
                # If after pruning the network, no new nodes have been removed,
                # the optimal solution won't have changed from previous iteration
                # so just reuse the solution
                if last_solution.network == network:
                    current_solution = last_solution.copy()
                    current_solution.threshold = threshold
                    solutions[threshold] = current_solution
                    last_solution = current_solution
                    current_network = network
                    continue
            wpl = weighted_laplacian_matrix(network)
            ident = np.eye(wpl.shape[0])

            # The network passed into LaplacianSmoothingModel will have its indices changed,
            # and will not match the ordering of the belongingness matrix, so make sure the
            # observed indices are aligned.
            lum = LaplacianSmoothingModel(
                network, self.normalized_belongingness_matrix[obs_ix, :], threshold,
                neighborhood_walker=self.neighborhood_walker,
                belongingness_normalization=renormalize_belongingness,
                variance_matrix=variance_matrix,
                inverse_variance_matrix=inverse_variance_matrix)
            updates = []
            taus = []
            lambda_values = np.arange(0.01, lambda_max, lambda_step)
            for lambd in lambda_values:
                if fit_tau:
                    tau = lum.estimate_tau_from_S0(rho, lambd)
                else:
                    tau = np.zeros(self.A0.shape[1])
                T = lum.optimize_observed_scores(lambd, lum.A0.dot(tau))
                A = ident + lambd * wpl

                # H = np.linalg.inv(A)
                H = fast_positive_definite_inverse(A)
                diag_H = np.diag(H)
                if len(diag_H) != len(T):
                    diag_H = diag_H[lum.obs_ix]
                    assert len(diag_H) == len(T)

                press_value = sum(
                    ((obs - T) / (1 - (diag_H - np.finfo(float).eps))) ** 2) / len(obs)
                press.append(press_value)
                updates.append(T)
                taus.append(tau)
            current_solution = NetworkTrimmingSearchSolution(
                threshold, lambda_values, np.array(press), network, np.array(obs),
                updates, taus, lum)

            solutions[threshold] = current_solution
            last_solution = current_solution
            current_network = network
        return solutions


def smooth_network(network, observed_compositions, threshold_step=0.5, apex_threshold=0.95,
                   belongingness_matrix=None, rho=DEFAULT_RHO, lambda_max=1,
                   include_missing=False, lmbda=None, model_state=None,
                   observation_aggregator=VariableObservationAggregation,
                   belongingness_normalization=NORMALIZATION, annotate_network=True):
    convert = GlycanCompositionSolutionRecord.from_chromatogram
    observed_compositions = [
        convert(o) for o in observed_compositions if _has_glycan_composition(o)]
    model = GlycomeModel(
        observed_compositions, network,
        belongingness_matrix=belongingness_matrix,
        observation_aggregator=observation_aggregator,
        belongingness_normalization=belongingness_normalization)
    # Deadlock?
    logger.info("... Begin Model Fitting")
    if model_state is None:
        reduction = model.find_threshold_and_lambda(
            rho=rho, threshold_step=threshold_step,
            lambda_max=lambda_max)
        if len(reduction) == 0:
            logger.info("... No Network Reduction Found")
            return None, None, None
        search = ThresholdSelectionGridSearch(model, reduction, apex_threshold)
        params = search.average_solution(lmbda=lmbda)
        if params is None:
            logger.info("... No Acceptable Solution. Could not fit model.")
            return None, None, None
    else:
        search = ThresholdSelectionGridSearch(model, None, apex_threshold)
        model_state.reindex(model)
        params = model_state
        if lmbda is not None:
            params.lmbda = lmbda
    if annotate_network:
        logger.info("... Projecting Solution Onto Network")
        annotated_network = search.annotate_network(params, include_missing=include_missing)
    else:
        annotated_network = None

    return annotated_network, search, params



from typing import Dict
import numpy as np
from scipy import linalg

from glycresoft.database.composition_network.graph import CompositionGraph

from .constants import DEFAULT_LAPLACIAN_REGULARIZATION
from .utils import fast_positive_definite_inverse

def adjacency_matrix(network):
    A = np.zeros((len(network), len(network)))
    for edge in network.edges:
        i, j = edge.node1.index, edge.node2.index
        A[i, j] = 1
        A[j, i] = 1
    for i in range(A.shape[0]):
        A[i, i] = 0
    return A


def weighted_adjacency_matrix(network):
    A = np.zeros((len(network), len(network)))
    A[:] = 1. / float('inf')
    for edge in network.edges:
        i, j = edge.node1.index, edge.node2.index
        A[i, j] = edge.weight
        A[j, i] = edge.weight
    for i in range(A.shape[0]):
        A[i, i] = 0
    return A


def degree_matrix(network):
    degrees = [len(n.edges) for n in network]
    return np.diag(degrees)


def weighted_degree_matrix(network):
    degrees = [sum(e.weight for e in n.edges) for n in network]
    return np.diag(degrees)


def laplacian_matrix(network):
    return degree_matrix(network) - adjacency_matrix(network)


def weighted_laplacian_matrix(network):
    return weighted_degree_matrix(network) - weighted_adjacency_matrix(network)


def assign_network(network, observed, copy=True):
    if copy:
        network = network.clone()
    solution_map = {}
    for case in observed:
        if case.glycan_composition is None:
            continue
        s = solution_map.get(case.glycan_composition)
        if s is None or s.score < case.score:
            solution_map[case.glycan_composition] = case

    for node in network.nodes:
        node.score = 0

    for composition, solution in solution_map.items():
        try:
            node = network[composition]
            node.score = solution.score
        except KeyError:
            # Not all exact compositions have nodes
            continue
    return network


def network_indices(network, threshold=0.0001):
    missing = []
    observed = []
    for node in network:
        if node.score < threshold:
            missing.append(node.index)
        else:
            observed.append(node.index)
    return np.array(observed, dtype=int), np.array(missing, dtype=int)


def make_blocks(network, observed, threshold=0.0001, regularize=DEFAULT_LAPLACIAN_REGULARIZATION):
    network = assign_network(network, observed)
    return BlockLaplacian(network, threshold, regularize)


def _reference_compute_missing_scores(blocks, observed_scores, t0=0., tm=0.):
    return -linalg.inv(blocks['mm']).dot(blocks['mo']).dot(observed_scores - t0) + tm


def _reference_optimize_observed_scores(blocks, lmbda, observed_scores, t0=0.):
    S = lmbda * (blocks["oo"] - blocks["om"].dot(linalg.inv(
                 blocks['mm'])).dot(blocks["mo"]))
    B = np.eye(len(observed_scores)) + S
    return linalg.inv(B).dot(observed_scores - t0) + t0


def scale_network(network, maximum):
    relmax = max([n.score for n in network.nodes])
    for node in network.nodes:
        node.score = node.score / relmax * maximum
    return network


def _key_from_memory(data: np.ndarray):
    return bytes(data.data)


class BlockLaplacian(object):
    regularize: float
    threshold: float

    obs_ix: np.ndarray
    miss_ix: np.ndarray

    matrix: np.ndarray
    blocks: Dict[str, np.ndarray]
    L_mm_inv: np.ndarray
    L_oo_inv: np.ndarray

    def __init__(self, network=None, threshold=0.0001, regularize=1.0):
        self.regularize = regularize
        self.threshold = threshold
        if network is not None:
            self._build_from_network(network)

    def _build_from_network(self, network: CompositionGraph):
        structure_matrix = weighted_laplacian_matrix(network)
        structure_matrix = structure_matrix + (np.eye(
            structure_matrix.shape[0]) * self.regularize)
        observed_indices, missing_indices = network_indices(network, self.threshold)

        self.obs_ix = observed_indices
        self.miss_ix = missing_indices

        self.matrix = structure_matrix
        # Compute Schur Complement
        if network.cache_state:
            key = _key_from_memory(self.obs_ix)
            if key in network.cache_state:
                self.blocks, self.L_mm_inv, self.L_oo_inv = network.cache_state[key]
            else:
                self._complement()
                network.cache_state[key] = self.blocks, self.L_mm_inv, self.L_oo_inv
        else:
            self._complement()

    def _blocks_from(self, matrix):
        oo_block = matrix[self.obs_ix, :][:, self.obs_ix]
        om_block = matrix[self.obs_ix, :][:, self.miss_ix]
        mo_block = matrix[self.miss_ix, :][:, self.obs_ix]
        mm_block = matrix[self.miss_ix, :][:, self.miss_ix]
        return {"oo": oo_block, "om": om_block, "mo": mo_block, "mm": mm_block}

    def _complement(self):
        '''Compute the Schur complement of the observed/unobserved blocked Laplacian
        matrix.
        '''
        self.blocks = self._blocks_from(self.matrix)
        self.L_mm_inv = fast_positive_definite_inverse(self['mm'])
        self.L_oo_inv = np.linalg.pinv(
            self["oo"] - (self['om'].dot(self.L_mm_inv).dot(self['mo'])))

    def __getitem__(self, k):
        return self.blocks[k]

    def __repr__(self):
        return "BlockLaplacian(%s)" % (', '.join({
            "%s: %r" % (k, v.shape) for k, v in self.blocks.items()
        }))



from uuid import uuid4

from glycresoft.serialize import (
    DatabaseBoundOperation, GlycopeptideHypothesis,
    Protein, Peptide, Glycopeptide, GlycanCombination,
    GlycanCombinationGlycanComposition, ProteinSite)

from glycresoft.serialize.utils import get_uri_for_instance

from glycresoft.database.builder.glycan.migrate import GlycanHypothesisMigrator

from glycresoft.task import TaskBase


class GlycopeptideHypothesisMigrator(DatabaseBoundOperation, TaskBase):
    def __init__(self, database_connection):
        DatabaseBoundOperation.__init__(self, database_connection)
        self.hypothesis_id = None
        self.glycan_hypothesis_id = None
        self._glycan_hypothesis_migrator = None
        self.protein_id_map = dict()
        self.peptide_id_map = dict()
        self.glycan_combination_id_map = dict()
        self.glycopeptide_id_map = dict()

    def clear(self):
        self.protein_id_map.clear()
        self.peptide_id_map.clear()
        self.glycan_combination_id_map.clear()
        self.glycopeptide_id_map.clear()

    def _migrate(self, obj):
        self.session.add(obj)
        self.session.flush()
        new_id = obj.id
        return new_id

    def commit(self):
        self.session.commit()

    def migrate_hypothesis(self, hypothesis):
        new_inst = GlycopeptideHypothesis(
            name=hypothesis.name,
            uuid=str(uuid4().hex),
            status=hypothesis.status,
            glycan_hypothesis_id=None,
            parameters=dict(hypothesis.parameters))
        new_inst.parameters["original_uuid"] = hypothesis.uuid
        new_inst.parameters['copied_from'] = get_uri_for_instance(hypothesis)
        self.hypothesis_id = self._migrate(new_inst)
        self.create_glycan_hypothesis_migrator(hypothesis.glycan_hypothesis)
        new_inst.glycan_hypothesis_id = self.glycan_hypothesis_id
        self._migrate(new_inst)

    def migrate_glycan_composition(self, glycan_composition):
        self._glycan_hypothesis_migrator.migrate_glycan_composition(
            glycan_composition)

    def migrate_glycan_combination(self, glycan_combination):
        new_inst = GlycanCombination(
            calculated_mass=glycan_combination.calculated_mass,
            formula=glycan_combination.formula,
            composition=glycan_combination.composition,
            count=glycan_combination.count,
            hypothesis_id=self.hypothesis_id)
        self.glycan_combination_id_map[glycan_combination.id] = self._migrate(new_inst)
        component_acc = []
        for component, count in glycan_combination.components.add_columns(
                GlycanCombinationGlycanComposition.c.count):
            new_component_id = self._glycan_hypothesis_migrator.glycan_composition_id_map[component.id]
            component_acc.append(
                {"glycan_id": new_component_id, "combination_id": new_inst.id, "count": count})
        if component_acc:
            self.session.execute(GlycanCombinationGlycanComposition.insert(), component_acc)

    def migrate_protein(self, protein):
        new_inst = Protein(
            name=protein.name,
            protein_sequence=protein.protein_sequence,
            other=dict(protein.other or dict()),
            hypothesis_id=self.hypothesis_id)
        new_id = self._migrate(new_inst)
        self.protein_id_map[protein.id] = new_id
        for position in protein.sites.all():
            self.session.add(
                ProteinSite(location=position.location, name=position.name, protein_id=new_id))
        self.session.flush()

    def create_glycan_hypothesis_migrator(self, glycan_hypothesis):
        self._glycan_hypothesis_migrator = GlycanHypothesisMigrator(self._original_connection)
        self._glycan_hypothesis_migrator.bridge(self)
        self._glycan_hypothesis_migrator.migrate_hypothesis(glycan_hypothesis)
        self.glycan_hypothesis_id = self._glycan_hypothesis_migrator.hypothesis_id

    def migrate_peptides_bulk(self, peptides):
        reverse_map = {}
        n = len(peptides)
        buffer = []
        for i, peptide in enumerate(peptides):
            if i % 15000 == 0 and i:
                self.log("... Migrating Peptides %d/%d (%0.2f%%)" %
                         (i, n, i * 100.0 / n))
                self.session.bulk_save_objects(buffer)
                buffer = []
            new_inst = Peptide(
                calculated_mass=peptide.calculated_mass,
                base_peptide_sequence=peptide.base_peptide_sequence,
                modified_peptide_sequence=peptide.modified_peptide_sequence,
                formula=peptide.formula,
                count_glycosylation_sites=peptide.count_glycosylation_sites,
                count_missed_cleavages=peptide.count_missed_cleavages,
                count_variable_modifications=peptide.count_variable_modifications,
                start_position=peptide.start_position,
                end_position=peptide.end_position,
                peptide_score=peptide.peptide_score,
                peptide_score_type=peptide.peptide_score_type,
                sequence_length=peptide.sequence_length,
                # Update the protein id
                protein_id=self.protein_id_map[peptide.protein_id],
                # Update the hypothesis id
                hypothesis_id=self.hypothesis_id,
                n_glycosylation_sites=peptide.n_glycosylation_sites,
                o_glycosylation_sites=peptide.o_glycosylation_sites,
                gagylation_sites=peptide.gagylation_sites)
            buffer.append(new_inst)
            reverse_map[(peptide.modified_peptide_sequence,
                         self.protein_id_map[peptide.protein_id],
                         peptide.start_position, )] = peptide.id

        if buffer:
            self.session.bulk_save_objects(buffer)
            buffer = []

        self.session.flush()
        self.log("... Reconstructing Reverse Mapping")
        migrated = self.session.query(
            Peptide.id, Peptide.modified_peptide_sequence,
            Peptide.protein_id, Peptide.start_position).filter(
                Peptide.hypothesis_id == self.hypothesis_id).yield_per(10000)
        m = 0
        for peptide in migrated:
            m += 1
            self.peptide_id_map[
                reverse_map[peptide.modified_peptide_sequence,
                            peptide.protein_id, peptide.start_position]] = peptide.id
        if m != n:
            raise ValueError("Peptide Migration Unsuccessful")
        return m

    def migrate_peptide(self, peptide):
        new_inst = Peptide(
            calculated_mass=peptide.calculated_mass,
            base_peptide_sequence=peptide.base_peptide_sequence,
            modified_peptide_sequence=peptide.modified_peptide_sequence,
            formula=peptide.formula,
            count_glycosylation_sites=peptide.count_glycosylation_sites,
            count_missed_cleavages=peptide.count_missed_cleavages,
            count_variable_modifications=peptide.count_variable_modifications,
            start_position=peptide.start_position,
            end_position=peptide.end_position,
            peptide_score=peptide.peptide_score,
            peptide_score_type=peptide.peptide_score_type,
            sequence_length=peptide.sequence_length,
            # Update the protein id
            protein_id=self.protein_id_map[peptide.protein_id],
            # Update the hypothesis id
            hypothesis_id=self.hypothesis_id,
            n_glycosylation_sites=peptide.n_glycosylation_sites,
            o_glycosylation_sites=peptide.o_glycosylation_sites,
            gagylation_sites=peptide.gagylation_sites)
        self.peptide_id_map[peptide.id] = self._migrate(new_inst)

    def migrate_glycopeptide_bulk(self, glycopeptides):
        reverse_map = {}
        n = len(glycopeptides)
        buffer = []
        for i, glycopeptide in enumerate(glycopeptides):
            if i % 50000 == 0 and i:
                self.log("... Migrating Glycopeptide %d/%d (%0.2f%%)" % (i, n, i * 100.0 / n))
                self.session.bulk_save_objects(buffer)
                buffer = []
            new_inst = Glycopeptide(
                calculated_mass=glycopeptide.calculated_mass,
                formula=glycopeptide.formula,
                glycopeptide_sequence=glycopeptide.glycopeptide_sequence,
                peptide_id=self.peptide_id_map[glycopeptide.peptide_id],
                protein_id=self.protein_id_map[glycopeptide.protein_id],
                hypothesis_id=self.hypothesis_id,
                glycan_combination_id=self.glycan_combination_id_map[glycopeptide.glycan_combination_id])
            buffer.append(new_inst)
            reverse_map[(glycopeptide.glycopeptide_sequence,
                         self.peptide_id_map[glycopeptide.peptide_id],
                         self.glycan_combination_id_map[glycopeptide.glycan_combination_id], )] = glycopeptide.id

        if buffer:
            self.session.bulk_save_objects(buffer)
            buffer = []

        self.session.flush()
        self.log("... Reconstructing Reverse Mapping")
        migrated = self.session.query(
            Glycopeptide.id, Glycopeptide.glycopeptide_sequence,
            Glycopeptide.peptide_id, Glycopeptide.glycan_combination_id).filter(
                Glycopeptide.hypothesis_id == self.hypothesis_id).yield_per(10000)

        m = 0
        for glycopeptide in migrated:
            m += 1
            self.glycopeptide_id_map[
                reverse_map[glycopeptide.glycopeptide_sequence,
                            glycopeptide.peptide_id,
                            glycopeptide.glycan_combination_id]] = glycopeptide.id
        if m != n:
            raise ValueError("Glycopeptide Migration Unsuccessful")
        return m

    def migrate_glycopeptide(self, glycopeptide):
        new_inst = Glycopeptide(
            calculated_mass=glycopeptide.calculated_mass,
            formula=glycopeptide.formula,
            glycopeptide_sequence=glycopeptide.glycopeptide_sequence,
            peptide_id=self.peptide_id_map[glycopeptide.peptide_id],
            protein_id=self.protein_id_map[glycopeptide.protein_id],
            hypothesis_id=self.hypothesis_id,
            glycan_combination_id=self.glycan_combination_id_map[glycopeptide.glycan_combination_id])
        self.glycopeptide_id_map[glycopeptide.id] = self._migrate(new_inst)



import os
import re
import operator
import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, DefaultDict, Dict, List, Mapping, Optional, Set, Tuple

from six import string_types as basestring

from glycopeptidepy.io import fasta as glycopeptidepy_fasta
from glycopeptidepy.structure import sequence, modification, residue
from glycopeptidepy.enzyme import Protease, enzyme_rules
from glypy.composition import formula

from glycresoft.serialize import (
    Peptide, Protein, DatabaseBoundOperation)

from glycresoft.task import TaskBase

from ..common import ProteinReversingMixin

from .mzid_parser import Parser
from .peptide_permutation import (
    ProteinDigestor, n_glycan_sequon_sites,
    o_glycan_sequon_sites, gag_sequon_sites, UniprotProteinAnnotator)
from .remove_duplicate_peptides import DeduplicatePeptides
from .share_peptides import PeptideSharer
from .fasta import FastaProteinSequenceResolver

logger = logging.getLogger("mzid")


PeptideSequence = sequence.PeptideSequence
Residue = residue.Residue
Modification = modification.Modification
ModificationNameResolutionError = modification.ModificationNameResolutionError
AnonymousModificationRule = modification.AnonymousModificationRule

GT = "greater"
LT = "lesser"

PROTEOMICS_SCORE = {
    "PEAKS:peptideScore": GT,
    "mascot:score": GT,
    "PEAKS:proteinScore": GT,
    "MS-GF:EValue": LT,
    "Byonic:Score": GT,
    "percolator:Q value": LT,
    r"X\!Tandem:expect": GT,
    "Morpheus:Morpheus score": GT,
    "MetaMorpheus:score": GT,
}


def score_comparator(score_type):
    try:
        preference = PROTEOMICS_SCORE[score_type]
        if preference == LT:
            return operator.lt
        else:
            return operator.gt
    except KeyError:
        raise KeyError("Don't know how to compare score of type %r" % score_type)


WHITELIST_GLYCOSITE_PTMS = [Modification("Deamidation"), Modification("HexNAc")]


class allset(object):

    def __contains__(self, x):
        return True


class ParameterizedProtease(Protease):
    def __init__(self, name, used_missed_cleavages=1, cleavage_start=None, cleavage_end=None):
        super(ParameterizedProtease, self).__init__(name, cleavage_start, cleavage_end)
        self.used_missed_cleavages = used_missed_cleavages


def resolve_database_url(url):
    if url.startswith("file://"):
        path = url.replace("file://", "")
        while path.startswith("/") and len(path) > 0:
            path = path[1:]
        if os.path.exists(path):
            if os.path.isfile(path):
                return path
            else:
                raise IOError("File URI %r points to a directory" % url)
        else:
            raise IOError("File URI %r does not exist on local system" % url)
    elif url.startswith("http"):
        return url
    elif os.path.exists(url):
        if os.path.isfile(url):
            return url
        else:
            raise IOError("File path %r points to a directory" % url)
    else:
        raise IOError("Cannot resolve URL %r" % url)


def protein_names(mzid_path, pattern=r'.*'):
    pattern = re.compile(pattern)
    parser = Parser(mzid_path, retrieve_refs=False,
                    iterative=True, build_id_cache=False, use_index=False)
    for protein in parser.iterfind(
            "DBSequence", retrieve_refs=False, recursive=False, iterative=True):
        name = protein['accession']
        if pattern.match(name):
            yield name


def remove_peptide_sequence_alterations(base_sequence, insert_sites, delete_sites):
    """
    Remove all the sequence insertions and deletions in order to reconstruct the
    original peptide sequence.

    Parameters
    ----------
    base_sequence : str
        The peptide sequence string which contains a combination
        of insertion and deletions.
    insert_sites : list
        A list of (position, None) pairs indicating the position of
        an amino acid insertion to be removed.
    delete_sites : list
        A list of (position, residue) pairs indicating the position
        and identity of amino acids that were deleted and need to be
        re-inserted.

    Returns
    -------
    str
    """
    sequence_copy = list(base_sequence)

    alteration_sites = insert_sites + delete_sites
    alteration_sites.sort()
    shift = 0
    for position, residue_ in alteration_sites:
        if residue_ is None:
            sequence_copy.pop(position - shift)
            shift += 1
        else:
            sequence_copy.insert(position - shift + 1, residue_)
            shift -= 1
    sequence_copy = ''.join(sequence_copy)
    return sequence_copy


class PeptideGroup(object):
    """Maps Protein ID to :class:`Peptide`, keeping track of the top scoring example
    for each :class:`Peptide` instance observed from the set of target Proteins

    Attributes
    ----------
    has_target_match: bool
        Description
    members: dict
        Mapping from Protein ID to :class:`Peptide`
    scores: defaultdict(list)
        Mapping from Protein ID to score
    """

    members: Dict
    scores: DefaultDict[str, List[float]]
    has_target_match: bool

    def __init__(self):
        self.members = dict()
        self.scores = defaultdict(list)
        self.has_target_match = False
        self._first = None
        self._last = None

    def __getitem__(self, key):
        return self.members[key]

    def __setitem__(self, key, value):
        self.members[key] = value
        if self._first is None:
            self._first = value
        self._last = value

    def clear(self):
        self.members.clear()
        self.has_target_match = False
        self._first = None
        self._last = None

    def _update_has_target_match(self, protein_set: Set[str]) -> bool:
        for key in self.members:
            if key in protein_set:
                return True
        return False

    def update_state(self, protein_set: Set[str]):
        had = self.has_target_match
        if not had:
            has = self.has_target_match = self._update_has_target_match(protein_set)
        else:
            has = had
        if not had and has:
            for key in list(self.members):
                if key not in protein_set:
                    self.members.pop(key)
                    self.scores.pop(key)
        else:
            if self._last.protein_id not in protein_set:
                try:
                    self.members.pop(self._last.protein_id)
                    self.scores.pop(self._last.protein_id)
                except KeyError:
                    pass

    def keys(self):
        return self.members.keys()

    def values(self):
        return self.members.values()

    def items(self):
        return self.members.items()

    def bind_scores(self):
        acc = []
        for key, peptide in self.members.items():
            peptide.scores = self.scores[key]
            acc.extend(self.scores[key])
        return acc

    def first(self):
        return self._first

    def __nonzero__(self):
        if self.members:
            return True
        else:
            return False

    def __bool__(self):
        if self.members:
            return True
        else:
            return False

    def __len__(self):
        return len(self.members)

    def add(self, peptide):
        comparator = score_comparator(peptide.peptide_score_type)
        if self:
            first = self.first()
            if comparator(peptide.peptide_score, first.peptide_score):
                self.clear()
                self[peptide.protein_id] = peptide
            elif first.peptide_score == peptide.peptide_score:
                self[peptide.protein_id] = peptide
        else:
            self[peptide.protein_id] = peptide
        self.scores[peptide.protein_id].append(peptide.peptide_score)


class PeptideCollection(object):
    store: DefaultDict[str, PeptideGroup]
    protein_set: Set[str]
    scores: List[float]

    def __init__(self, protein_set):
        self.store = defaultdict(PeptideGroup)
        self.protein_set = protein_set
        self.scores = []

    def bind_scores(self):
        acc = []
        for key, value in self.store.items():
            acc.extend(value.bind_scores())
        self.scores = acc
        return acc

    def add(self, peptide):
        group = self.store[peptide.modified_peptide_sequence]
        group.add(peptide)
        group.update_state(self.protein_set)
        self.store[peptide.modified_peptide_sequence] = group

    def items(self):
        for key, mapping in self.store.items():
            yield key, mapping.values()

    def __len__(self):
        return sum(map(len, self.store.values()))


class MzIdentMLPeptide(object):
    peptide_dict: Dict[str, Any]
    insert_sites: List[Tuple[int, str]]
    deleteion_sites: List[Tuple[int, str]]

    modification_counter: int
    missed_cleavages: Optional[int]

    base_sequence: str
    peptide_sequence: PeptideSequence
    glycosite_candidates: List[int]

    modification_translation_table: Dict[str, Modification]
    enzyme: Optional[Protease]
    mzid_id: Optional[str]

    def __init__(self, peptide_dict, enzyme=None, constant_modifications=None,
                 modification_translation_table=None, process=True):
        if modification_translation_table is None:
            modification_translation_table = dict()
        if constant_modifications is None:
            constant_modifications = list()

        self.peptide_dict = peptide_dict

        self.insert_sites = []
        self.deleteion_sites = []
        self.modification_counter = 0
        self.missed_cleavages = 0

        self.base_sequence = peptide_dict["PeptideSequence"]
        self.peptide_sequence = PeptideSequence(peptide_dict["PeptideSequence"])

        self.glycosite_candidates = sequence.find_n_glycosylation_sequons(
            self.peptide_sequence, WHITELIST_GLYCOSITE_PTMS)

        self.constant_modifications = constant_modifications
        self.modification_translation_table = modification_translation_table
        self.enzyme = enzyme
        self.mzid_id = peptide_dict.get('id')

        if process:
            self.process()

    def handle_missed_cleavages(self):
        if self.enzyme is not None:
            self.missed_cleavages = self.enzyme.missed_cleavages(self.base_sequence)
        else:
            self.missed_cleavages = None

    def handle_substitutions(self):
        if "SubstitutionModification" in self.peptide_dict:
            subs = self.peptide_dict["SubstitutionModification"]
            for sub in subs:
                pos = sub['location'] - 1
                replace = Residue(sub["replacementResidue"])
                self.peptide_sequence.substitute(pos, replace)
                self.modification_counter += 1

    def add_modification(self, modification: Modification, position: int):
        if position == -1:
            targets = modification.rule.n_term_targets
            for t in targets:
                if t.position_modifier is not None and t.amino_acid_targets is None:
                    break
            else:
                position += 1
        if position == len(self.peptide_sequence):
            targets = modification.rule.c_term_targets
            for t in targets:
                if t.position_modifier is not None and t.amino_acid_targets is None:
                    break
            else:
                position -= 1
        if position == -1:
            self.peptide_sequence.n_term = modification
        elif position == len(self.peptide_sequence):
            self.peptide_sequence.c_term = modification
        else:
            self.peptide_sequence.add_modification(position, modification)

    def add_insertion(self, site: int, symbol=None):
        self.insert_sites.append((site, symbol))

    def add_deletion(self, site: int, symbol):
        self.deleteion_sites.append((site, symbol))

    def handle_modifications(self):
        if "Modification" in self.peptide_dict:
            mods: List[Dict[str, Any]] = self.peptide_dict["Modification"]
            for mod in mods:
                pos: int = mod["location"] - 1
                accession = None
                try:
                    if "unknown modification" in mod:
                        try:
                            _name = mod['unknown modification']
                            if _name in self.modification_translation_table:
                                modification = self.modification_translation_table[_name]()
                            else:
                                modification = Modification(str(_name))
                        except ModificationNameResolutionError:
                            raise KeyError("Cannot find key in %r" % (mod,))
                    else:
                        try:
                            _name = mod["name"]
                            accession = getattr(_name, "accession", None)
                            _name = str(_name)
                            if accession is not None:
                                accession = str(accession)
                                try:
                                    modification = Modification(accession)
                                except ModificationNameResolutionError:
                                    modification = Modification(_name)
                            else:
                                modification = Modification(_name)

                        except (KeyError, ModificationNameResolutionError) as e:
                            raise KeyError("Cannot find key %s in %r" % (e, mod))

                    try:
                        rule_text = "%s (%s)" % (_name, mod["residues"][0])
                        if (rule_text not in self.constant_modifications) and not (
                                pos in self.glycosite_candidates and modification in WHITELIST_GLYCOSITE_PTMS):
                            self.modification_counter += 1
                    except KeyError:
                        self.modification_counter += 1

                    self.add_modification(modification, pos)
                except KeyError:
                    if "unknown modification" in mod:
                        mod_description = mod["unknown modification"]
                        insertion = re.search(r"(\S{3})\sinsertion", mod_description)
                        deletion = re.search(r"(\S{3})\sdeletion", mod_description)
                        self.modification_counter += 1
                        if insertion:
                            self.add_insertion(mod['location'] - 1, None)
                        elif deletion:
                            sym = Residue(deletion.groups()[0]).symbol
                            self.add_deletion(mod['location'] - 1, sym)
                        elif 'monoisotopicMassDelta' in mod:
                            mass = float(mod['monoisotopicMassDelta'])
                            modification = AnonymousModificationRule(str(_name), mass)()
                            self.add_modification(modification, pos)
                        else:
                            raise
                    else:
                        raise

    def process(self):
        self.handle_substitutions()
        self.handle_modifications()
        self.handle_missed_cleavages()

    def original_sequence(self):
        sequence_copy = remove_peptide_sequence_alterations(
            self.base_sequence, self.insert_sites, self.deleteion_sites)
        return sequence_copy

    def __repr__(self):
        return f"{self.__class__.__name__}({self.peptide_sequence}, {self.glycosite_candidates})"


class _EvidenceMixin:
    @property
    def score_type(self) -> Tuple[Optional[float], Optional[str]]:
        score = score_type = None
        for k, v in self.peptide_dict.items():
            if k in PROTEOMICS_SCORE:
                score_type = k
                score = v
                break
        return score, score_type

    @property
    def evidence_list(self) -> List[Dict[str, Any]]:
        evidence_list = self.peptide_dict["PeptideEvidenceRef"]
        # Flatten the evidence list if it has extra nesting because of alternative
        # mzid parsing
        if isinstance(evidence_list[0], list):
            evidence_list = [x for sub in evidence_list for x in sub]
        return evidence_list


class PeptideIdentification(MzIdentMLPeptide, _EvidenceMixin):
    def __init__(self, peptide_dict, enzyme=None, constant_modifications=None,
                 modification_translation_table=None, process=True):
        super(PeptideIdentification, self).__init__(
            peptide_dict, enzyme, constant_modifications, modification_translation_table,
            process)


class ProteinStub(object):
    name: str
    id: int
    protein_sequence: str
    n_glycan_sequon_sites: List[int]
    o_glycan_sequon_sites: List[int]
    glycosaminoglycan_sequon_sites: List[int]

    def __len__(self):
        return len(self.protein_sequence)

    def __init__(self, name, id, sequence, n_glycan_sequon_sites, o_glycan_sequon_sites,
                 glycosaminoglycan_sequon_sites):
        self.name = name
        self.id = id
        self.protein_sequence = sequence
        self.n_glycan_sequon_sites = n_glycan_sequon_sites
        self.o_glycan_sequon_sites = o_glycan_sequon_sites
        self.glycosaminoglycan_sequon_sites = glycosaminoglycan_sequon_sites

    @classmethod
    def from_protein(cls, protein):
        return cls(
            str(protein.name),
            getattr(protein, 'id', None),
            str(protein),
            protein.n_glycan_sequon_sites,
            protein.o_glycan_sequon_sites,
            protein.glycosaminoglycan_sequon_sites)

    def __repr__(self):
        trailer = self.protein_sequence[:10] + '...'
        return f"{self.__class__.__name__}({self.name!r}, {self.id}, {trailer!r})"


class ProteinStubLoaderBase(object):
    store: Mapping[str, ProteinStub]

    def __init__(self, store=None):
        if store is None:
            store = dict()
        self.store = store

    def __getitem__(self, key: str) -> ProteinStub:
        return self.get_protein_by_name(key)

    def __contains__(self, key: str):
        return key in self.store

    def _load_protein(self, protein_name: str):
        raise NotImplementedError()

    def get_protein_by_name(self, protein_name: str) -> ProteinStub:
        try:
            return self.store[protein_name]
        except KeyError:
            protein = self._load_protein(protein_name)
            try:
                stub = ProteinStub.from_protein(protein)
            except AttributeError:
                logger.error("Failed to load stub for %s" % (protein_name,))
                raise
            self.store[protein_name] = stub
            return stub


class MemoryProteinStubLoader(ProteinStubLoaderBase):
    def _load_protein(self, protein_name):
        raise KeyError(protein_name)


class DatabaseProteinStubLoader(ProteinStubLoaderBase):
    hypothesis_id: int
    alias_map: Dict[str, str]

    def __init__(self, session, hypothesis_id, store=None, alias_map: Optional[Dict[str, str]]=None):
        if alias_map is None:
            alias_map = {}
        self.session = session
        self.hypothesis_id = hypothesis_id
        self.alias_map = alias_map
        super(DatabaseProteinStubLoader, self).__init__(store)

    def _load_protein(self, protein_name: str) -> Protein:
        if protein_name in self.alias_map:
            protein_name = self.alias_map[protein_name]
        protein = self.session.query(Protein).filter(
            Protein.name == protein_name,
            Protein.hypothesis_id == self.hypothesis_id).first()
        if protein is None:
            breakpoint()
            raise KeyError(protein_name)
        return protein


class FastaProteinStubLoader(ProteinStubLoaderBase):
    def __init__(self, fasta_path, store=None):
        super(FastaProteinStubLoader, self).__init__(store)
        self.parser = glycopeptidepy_fasta.ProteinFastaFileParser(fasta_path, index=True)

    def _load_protein(self, protein_name: str):
        return self.parser[protein_name]


class PeptideMapperBase(object):
    enzyme: Protease
    constant_modifications: Dict[str, Modification]
    modification_translation_table: Dict[str, Modification]

    peptide_grouper: PeptideCollection
    protein_loader: ProteinStubLoaderBase


    counter: int
    peptide_length_range: Tuple[int, int]


    def __init__(self, enzyme=None, constant_modifications=None, modification_translation_table=None,
                 protein_filter=None, peptide_length_range=(5, 60)):
        if protein_filter is None:
            protein_filter = allset()

        self.enzyme = enzyme
        self.constant_modifications = constant_modifications or []
        self.modification_translation_table = modification_translation_table or {}

        self.peptide_grouper = PeptideCollection(protein_filter)
        self.counter = 0

        self.protein_loader = self._make_protein_loader()
        self.peptide_length_range = peptide_length_range

    def _make_protein_loader(self):
        raise NotImplementedError()


class PeptideConverter(PeptideMapperBase):
    hypothesis_id: int
    include_cysteine_n_glycosylation: bool

    def __init__(self, session, hypothesis_id, enzyme=None, constant_modifications=None,
                 modification_translation_table=None, protein_filter=None,
                 peptide_length_range=(5, 60),
                 include_cysteine_n_glycosylation: bool = False,
                 protein_alias_map: Optional[Dict[str, str]]=None):
        self.session = session
        self.hypothesis_id = hypothesis_id
        self.include_cysteine_n_glycosylation = include_cysteine_n_glycosylation
        self.protein_alias_map = protein_alias_map
        super(PeptideConverter, self).__init__(
            enzyme, constant_modifications, modification_translation_table,
            protein_filter, peptide_length_range)

    def _make_protein_loader(self):
        return DatabaseProteinStubLoader(self.session, self.hypothesis_id, alias_map=self.protein_alias_map)

    def get_protein(self, evidence):
        return self.protein_loader[evidence['accession']]

    def pack_peptide(self, peptide_ident: PeptideIdentification, start: int, end: int, score: float,
                     score_type: str, parent_protein: Protein):
        match = Peptide(
            calculated_mass=peptide_ident.peptide_sequence.mass,
            base_peptide_sequence=peptide_ident.base_sequence,
            modified_peptide_sequence=str(peptide_ident.peptide_sequence),
            formula=formula(peptide_ident.peptide_sequence.total_composition()),
            count_glycosylation_sites=None,
            count_missed_cleavages=peptide_ident.missed_cleavages,
            count_variable_modifications=peptide_ident.modification_counter,
            start_position=start,
            end_position=end,
            peptide_score=score,
            peptide_score_type=score_type,
            sequence_length=end - start,
            protein_id=parent_protein.id,
            hypothesis_id=self.hypothesis_id)
        n_glycosites = n_glycan_sequon_sites(
            match, parent_protein, include_cysteine=self.include_cysteine_n_glycosylation)
        o_glycosites = o_glycan_sequon_sites(match, parent_protein)
        gag_glycosites = gag_sequon_sites(match, parent_protein)
        match.count_glycosylation_sites = len(n_glycosites) + len(o_glycosites)
        match.n_glycosylation_sites = sorted(n_glycosites)
        match.o_glycosylation_sites = sorted(o_glycosites)
        match.gagylation_sites = sorted(gag_glycosites)
        return match

    def copy_db_peptide(self, db_peptide: Peptide):
        dup = Peptide(
            calculated_mass=db_peptide.calculated_mass,
            base_peptide_sequence=db_peptide.base_peptide_sequence,
            modified_peptide_sequence=db_peptide.modified_peptide_sequence,
            formula=db_peptide.formula,
            count_glycosylation_sites=db_peptide.count_glycosylation_sites,
            count_missed_cleavages=db_peptide.count_missed_cleavages,
            count_variable_modifications=db_peptide.count_variable_modifications,
            start_position=db_peptide.start_position,
            end_position=db_peptide.end_position,
            peptide_score=db_peptide.peptide_score,
            peptide_score_type=db_peptide.peptide_score_type,
            sequence_length=db_peptide.sequence_length,
            protein_id=db_peptide.protein_id,
            hypothesis_id=db_peptide.hypothesis_id,
            n_glycosylation_sites=db_peptide.n_glycosylation_sites,
            o_glycosylation_sites=db_peptide.o_glycosylation_sites,
            gagylation_sites=db_peptide.gagylation_sites)
        return dup

    def has_occupied_glycosites(self, db_peptide: Peptide):
        occupied_sites = []
        n_glycosylation_sites = db_peptide.n_glycosylation_sites
        peptide_obj = PeptideSequence(db_peptide.modified_peptide_sequence)

        for site in n_glycosylation_sites:
            if peptide_obj[site][1]:
                occupied_sites.append(site)
        return len(occupied_sites) > 0

    def clear_sites(self, db_peptide: Peptide):
        # TODO: Make this a combinatorial generator so that it optionally clears each
        # putative combination of glycosites across N/O forms.
        occupied_sites = []
        n_glycosylation_sites = db_peptide.n_glycosylation_sites
        peptide_obj = PeptideSequence(db_peptide.modified_peptide_sequence)

        for site in n_glycosylation_sites:
            if peptide_obj[site][1]:
                occupied_sites.append(site)

        for site in occupied_sites:
            peptide_obj.drop_modification(site, peptide_obj[site][1][0])

        copy = self.copy_db_peptide(db_peptide)
        copy.calculated_mass = peptide_obj.mass
        copy.modified_peptide_sequence = str(peptide_obj)
        copy.count_variable_modifications -= len(occupied_sites)
        return copy

    def sequence_starts_at(self, sequence: str, parent_protein: Protein) -> int:
        found = parent_protein.protein_sequence.find(sequence)
        if found == -1:
            raise ValueError("Peptide not found in Protein\n%s\n%r\n\n" % (parent_protein.name, sequence))
        return found

    def handle_peptide_dict(self, peptide_dict: Dict[str, Any]):
        peptide_ident = PeptideIdentification(
            peptide_dict, self.enzyme, self.constant_modifications,
            self.modification_translation_table)

        score, score_type = peptide_ident.score_type
        evidence_list = peptide_ident.evidence_list

        for evidence in evidence_list:
            if "skip" in evidence:
                continue
            if evidence.get("isDecoy", False):
                continue

            parent_protein = self.get_protein(evidence)
            if parent_protein is None:
                continue

            start = evidence["start"] - 1
            end = evidence["end"]
            length = len(peptide_ident.base_sequence)
            if not (self.peptide_length_range[0] <= length <= self.peptide_length_range[1]):
                continue

            sequence_copy = peptide_ident.original_sequence()
            found = self.sequence_starts_at(sequence_copy, parent_protein)

            if found != start:
                start = found
                end = start + length

            match = self.pack_peptide(
                peptide_ident, start, end, score, score_type, parent_protein)
            self.add_to_group(match)
            if self.has_occupied_glycosites(match):
                cleared = self.clear_sites(match)
                self.add_to_group(cleared)

    def add_to_group(self, match):
        self.counter += 1
        self.peptide_grouper.add(match)

    def save_accumulation(self):
        acc = []
        self.peptide_grouper.bind_scores()
        for key, group in self.peptide_grouper.items():
            acc.extend(group)
        self.session.bulk_save_objects(acc)
        self.session.commit()


class MzIdentMLProteomeExtraction(TaskBase):
    def __init__(self, mzid_path, reference_fasta=None, peptide_length_range=(5, 60)):
        self.mzid_path = mzid_path
        self.reference_fasta = reference_fasta
        self.parser = Parser(mzid_path, retrieve_refs=True, iterative=True, use_index=True)
        self.enzymes = []
        self.constant_modifications = []
        self.modification_translation_table = {}

        self._protein_resolver = None
        self._ignore_protein_regex = None
        self._used_database_path = None

        self.peptide_length_range = peptide_length_range or (5, 60)

    def load_enzyme(self):
        self.parser.reset()
        # self.enzymes = list({e['name'].lower() for e in self.parser.iterfind(
        #     "EnzymeName", retrieve_refs=True, iterative=True)})
        enzymes = list(self.parser.iterfind("Enzyme", retrieve_refs=True, iterative=True))
        processed_enzymes = []
        for enz in enzymes:
            permitted_missed_cleavages = int(enz.get("missedCleavages", 1))
            # It's a standard enzyme, so we can look it up by name
            if "EnzymeName" in enz:
                try:
                    enz_name = enz.get('EnzymeName')
                    if isinstance(enz_name, dict):
                        if len(enz_name) == 1:
                            enz_name = list(enz_name)[0]
                        else:
                            self.log("Could not interpret Enzyme Name %r" % (enz,))
                    protease = ParameterizedProtease(enz_name.lower(), permitted_missed_cleavages)
                    processed_enzymes.append(protease)
                except (KeyError, re.error) as e:
                    self.log("Could not resolve protease from name %r (%s)" % (enz['EnzymeName'].lower(), e))
            elif "SiteRegexp" in enz:
                pattern = enz['SiteRegexp']
                try:
                    protease = ParameterizedProtease(pattern, permitted_missed_cleavages)
                    processed_enzymes.append(protease)
                except re.error as e:
                    self.log("Could not resolve protease from name %r (%s)" % (enz['SiteRegexp'].lower(), e))
                except KeyError:
                    self.log("No protease information available: %r" % (enz,))
            elif "name" in enz and enz['name'].lower() in enzyme_rules:
                protease = ParameterizedProtease(enz["name"].lower(), permitted_missed_cleavages)
                processed_enzymes.append(protease)
            elif "id" in enz and enz['id'].lower() in enzyme_rules:
                protease = ParameterizedProtease(enz["id"].lower(), permitted_missed_cleavages)
                processed_enzymes.append(protease)
            else:
                self.log("No protease information available: %r" % (enz,))
        self.enzymes = list(set(processed_enzymes))

    def load_modifications(self):
        self.parser.reset()
        search_param_modifications = list(self.parser.iterfind(
            "ModificationParams", retrieve_refs=True, iterative=True))
        constant_modifications = []

        for param in search_param_modifications:
            for mod in param['SearchModification']:
                try:
                    name = mod['name']
                except KeyError:
                    name = mod['unknown modification']
                    try:
                        Modification(str(name))
                    except ModificationNameResolutionError:
                        self.modification_translation_table[name] = modification.AnonymousModificationRule(
                            str(name), mod['massDelta'])

                residues = mod['residues']
                if mod.get('fixedMod', False):
                    identifier = "%s (%s)" % (name, ''.join(residues).replace(" ", ""))
                    constant_modifications.append(identifier)
        self.constant_modifications = constant_modifications

    def _make_protein_resolver(self):
        if self.reference_fasta is not None:
            self._protein_resolver = FastaProteinSequenceResolver(self.reference_fasta)
            return
        else:
            path = self._find_used_database()
            if path is not None:
                self._protein_resolver = FastaProteinSequenceResolver(path)
                return
        raise ValueError("Cannot construct a Protein Resolver. Cannot fetch additional protein information.")

    def _clear_protein_resolver(self):
        self._protein_resolver = None

    def _find_used_database(self):
        if self._used_database_path is not None:
            return self._used_database_path
        self.parser.reset()
        databases = list(self.parser.iterfind("SearchDatabase", iterative=True))
        # use only the first database
        if len(databases) > 1:
            self.log("... %d databases found: %r" % (len(databases), databases))
            self.log("... Using first only")
        database = databases[0]
        if "decoy DB accession regexp" in database:
            self._ignore_protein_regex = re.compile(database["decoy DB accession regexp"])
        if "FASTA format" in database.get('FileFormat', {}):
            self.log("... Database described in FASTA format")
            db_location = database.get("location")
            if db_location is None:
                raise ValueError("No location present for database")
            else:
                try:
                    path = resolve_database_url(db_location)
                    with open(path, 'r') as handle:
                        for i, line in enumerate(handle):
                            if i > 1000:
                                raise ValueError("No FASTA Header before thousandth line. Probably not a FASTA file")
                            if line.startswith(">"):
                                break
                    self._used_database_path = path
                    return path
                except (IOError, ValueError):
                    return None
        else:
            return None

    def resolve_protein(self, name):
        if self._protein_resolver is None:
            self._make_protein_resolver()
        proteins = self._protein_resolver.find(name)
        if len(proteins) > 1:
            self.log("Protein Name %r resolved to multiple proteins: %r. Using first only." % (name, proteins))
        elif len(proteins) == 0:
            raise KeyError(name)
        return proteins[0]

    def _load_peptides(self):
        self.parser.reset()
        i = 0
        try:
            enzyme = self.enzymes[0]
        except IndexError as e:
            logger.exception("Enzyme not found.", exc_info=e)
            enzyme = None
        for peptide in self.parser.iterfind("Peptide", iterative=True, retrieve_refs=True):
            i += 1
            mzid_peptide = MzIdentMLPeptide(
                peptide, enzyme, self.constant_modifications,
                self.modification_translation_table)
            if i % 1000 == 0:
                self.log("Loaded Peptide %r" % (mzid_peptide,))
            yield mzid_peptide

    def _map_peptide_to_proteins(self):
        self.parser.reset()
        i = 0
        peptide_to_proteins = defaultdict(set)
        for evidence in self.parser.iterfind('PeptideEvidence', iterative=True):
            i += 1
            peptide_to_proteins[evidence['peptide_ref']].add(
                evidence['dBSequence_ref'])
        return peptide_to_proteins

    def _handle_protein(self, name, sequence, data):
        raise NotImplementedError()

    def load_proteins(self):
        self.parser.reset()
        self._find_used_database()
        protein_map = {}
        self.parser.reset()
        for protein in self.parser.iterfind("DBSequence", recursive=True, iterative=True):
            seq = protein.pop('Seq', None)
            name = protein.pop('accession')
            if seq is None:
                try:
                    prot = self.resolve_protein(name)
                    seq = prot.protein_sequence
                except KeyError:
                    if self._can_ignore_protein(name):
                        continue
                    else:
                        self.log("Could not resolve protein %r" % (name,))

            if "protein description" in protein:
                name = " ".join(
                    (name,
                     protein.pop("protein description"))
                )

            if name in protein_map:
                if seq != protein_map[name].protein_sequence:
                    self.log("Multiple proteins with the name %r" % name)
                continue
            try:
                p = self._handle_protein(name, seq, protein)
                protein_map[name] = p
            except residue.UnknownAminoAcidException:
                self.log("Unknown Amino Acid in %r" % (name,))
                continue
            except Exception as e:
                self.log("%r skipped: %r" % (name, e))
                continue
        self._clear_protein_resolver()
        return protein_map


class Proteome(DatabaseBoundOperation, MzIdentMLProteomeExtraction):
    def __init__(self, mzid_path, connection, hypothesis_id, include_baseline_peptides=True,
                 target_proteins=None, reference_fasta=None,
                 peptide_length_range=(5, 60), use_uniprot=True,
                 include_cysteine_n_glycosylation: bool=False,
                 uniprot_source_file: Optional[str]=None):
        DatabaseBoundOperation.__init__(self, connection)
        MzIdentMLProteomeExtraction.__init__(self, mzid_path, reference_fasta)
        if target_proteins is None:
            target_proteins = []
        self.hypothesis_id = hypothesis_id
        self.target_proteins = target_proteins
        self.include_baseline_peptides = include_baseline_peptides
        self.peptide_length_range = peptide_length_range or (5, 60)
        self.use_uniprot = use_uniprot
        self.accession_map = {}
        self.include_cysteine_n_glycosylation = include_cysteine_n_glycosylation
        self.uniprot_source_file = uniprot_source_file

    def _can_ignore_protein(self, name):
        if name not in self.target_proteins:
            return True
        elif (self._ignore_protein_regex is not None) and (
                self._ignore_protein_regex.match(name)):
            return True
        return False

    def _make_protein_from_state(self, name: str, seq: str, state: dict):
        p = Protein(
            name=name,
            protein_sequence=seq,
            other=state,
            hypothesis_id=self.hypothesis_id)
        p._init_sites(self.include_cysteine_n_glycosylation)
        return p

    def load_proteins(self):
        self.parser.reset()
        self._find_used_database()
        session = self.session
        protein_map = {}
        self.accession_map = {}
        self.parser.reset()
        for protein in self.parser.iterfind(
                "DBSequence", retrieve_refs=True, recursive=True, iterative=True):
            seq = protein.pop('Seq', None)
            accession = name = protein.pop('accession')
            if seq is None:
                try:
                    prot = self.resolve_protein(name)
                    seq = prot.protein_sequence
                except KeyError:
                    if self._can_ignore_protein(name):
                        continue
                    else:
                        self.log("Could not resolve protein %r" % (name,))

            if "protein description" in protein:
                name = " ".join(
                    (name,
                     protein.pop("protein description"))
                )

            if name in protein_map:
                if seq != protein_map[name].protein_sequence:
                    self.log("Multiple proteins with the name %r" % name)
                continue
            try:
                p = self._make_protein_from_state(name, seq, protein)
                session.add(p)
                session.flush()
                protein_map[name] = p
                self.accession_map[accession] = name
            except residue.UnknownAminoAcidException:
                self.log("Unknown Amino Acid in %r" % (name,))
                continue
            except Exception as e:
                self.log("%r skipped: %r" % (name, e))
                continue
        session.commit()
        self._clear_protein_resolver()

    def _make_peptide_converter(self, session, protein_filter: Set[int], enzyme: Protease, **kwargs):
        return PeptideConverter(
            session, self.hypothesis_id, enzyme,
            self.constant_modifications,
            self.modification_translation_table,
            protein_filter=protein_filter,
            protein_alias_map=self.accession_map, **kwargs)

    def load_spectrum_matches(self):
        self.parser.reset()
        last = 0
        i = 0
        try:
            enzyme = self.enzymes[0]
        except IndexError as e:
            logger.exception("Enzyme not found.", exc_info=e)
            enzyme = None
        session = self.session

        protein_filter = set(self.retrieve_target_protein_ids())

        peptide_converter = self._make_peptide_converter(
            session, protein_filter, enzyme)

        for spectrum_identification in self.parser.iterfind(
                "SpectrumIdentificationItem", retrieve_refs=True, iterative=True):
            peptide_converter.handle_peptide_dict(spectrum_identification)
            i += 1
            if i % 1000 == 0:
                self.log("... %d spectrum matches processed." % i)
            if (peptide_converter.counter - last) > 1000000:
                last = peptide_converter.counter
                self.log("... %d peptides saved. %d distinct cases." % (
                    peptide_converter.counter, len(
                        peptide_converter.peptide_grouper)))
        self.log("... %d peptides saved. %d distinct cases." % (
            peptide_converter.counter, len(
                peptide_converter.peptide_grouper)))
        peptide_converter.save_accumulation()

    def load(self):
        self.log("... Loading Enzyme")
        self.load_enzyme()
        self.log("... Loading Modifications")
        self.load_modifications()
        self.log("... Loading Proteins")
        self.load_proteins()
        self.log("... Loading Spectrum Matches")
        self.load_spectrum_matches()
        self.log("Sharing Common Peptides")
        self.remove_duplicates()
        self.share_common_peptides()
        self.remove_duplicates()
        self.log("... %d Peptides Total" % (self.count_peptides()))
        if self.include_baseline_peptides:
            self.log("... Building Baseline Peptides")
            self.build_baseline_peptides()
            self.remove_duplicates()
            self.log("... %d Peptides Total" % (self.count_peptides()))
            if self.use_uniprot:
                self.split_proteins()
        self.log("... Removing Duplicate Peptides")
        self.remove_duplicates()
        self.log("... %d Peptides Total" % (self.count_peptides()))

    def retrieve_target_protein_ids(self):
        if not self.target_proteins:
            return [
                i[0] for i in
                self.query(Protein.id).filter(
                    Protein.hypothesis_id == self.hypothesis_id).all()
            ]
        else:
            result = []
            for target in self.target_proteins:
                if isinstance(target, basestring):
                    if target in self.accession_map:
                        target = self.accession_map[target]
                    match = self.query(Protein.id).filter(
                        Protein.name == target,
                        Protein.hypothesis_id == self.hypothesis_id).first()
                    if match:
                        result.append(match[0])
                    else:
                        self.log("Could not locate protein '%s'" % target)
                elif isinstance(target, int):
                    result.append(target)
            return result

    def get_target_proteins(self):
        ids = self.retrieve_target_protein_ids()
        return [self.session.query(Protein).get(i) for i in ids]

    def count_peptides(self):
        peptide_count = self.session.query(Peptide).filter(
            Peptide.hypothesis_id == self.hypothesis_id).count()
        return peptide_count

    def make_restricted_modification_rules(self, rule_strings: List[str]):
        standard_rules = []
        alternative_rules: DefaultDict[str, List[str]] = DefaultDict(list)
        for rule_str in rule_strings:
            name = modification.RestrictedModificationTable.extract_name(rule_str)
            if name in self.modification_translation_table:
                if isinstance(self.modification_translation_table[name], AnonymousModificationRule):
                    alternative_rules[name].append(rule_str)
                else:
                    standard_rules.append(rule_str)
            else:
                standard_rules.append(rule_str)

        mod_table = modification.RestrictedModificationTable(
            constant_modifications=standard_rules,
            variable_modifications=[])
        rules = [mod_table[c] for c in standard_rules]
        if alternative_rules:
            for name, rule_strs in alternative_rules.items():
                rule: AnonymousModificationRule = self.modification_translation_table[name].clone()
                rule.targets = {modification.RestrictedModificationTable.extract_target(s) for s in rule_strs}
                rules.append(rule)
        return rules

    def build_baseline_peptides(self):
        const_modifications = self.make_restricted_modification_rules(self.constant_modifications)
        digestor = ProteinDigestor(
            self.enzymes[0], const_modifications,
            max_missed_cleavages=self.enzymes[0].used_missed_cleavages,
            include_cysteine_n_glycosylation=self.include_cysteine_n_glycosylation,
            min_length=self.peptide_length_range[0],
            max_length=self.peptide_length_range[1])
        accumulator = []
        i = 0
        for protein in self.get_target_proteins():
            for peptide in digestor.process_protein(protein):
                peptide.hypothesis_id = self.hypothesis_id
                accumulator.append(peptide)
                i += 1
                if len(accumulator) > 5000:
                    self.session.bulk_save_objects(accumulator)
                    self.session.commit()
                    accumulator = []
                if i % 1000 == 0:
                    self.log("... %d Baseline Peptides Created" % i)

        self.session.bulk_save_objects(accumulator)
        self.session.commit()

    def share_common_peptides(self):
        sharer = PeptideSharer(self._original_connection, self.hypothesis_id)
        proteins = self.get_target_proteins()
        i = 0
        n = len(proteins)
        for protein in proteins:
            i += 1
            # self.log("... Accumulating Proteins for %r" % protein)
            sharer.find_contained_peptides(protein)
            if i % 5 == 0:
                self.log("... %0.3f%% Done (%s)" % (i / float(n) * 100., protein.name))

    def split_proteins(self):
        const_modifications = self.make_restricted_modification_rules(self.constant_modifications)
        protein_ids = self.retrieve_target_protein_ids()
        if self.use_uniprot:
            annotator = UniprotProteinAnnotator(
                self,
                protein_ids,
                const_modifications,
                [],
                uniprot_source_file=self.uniprot_source_file
            )
            annotator.run()

    def remove_duplicates(self):
        DeduplicatePeptides(self._original_connection, self.hypothesis_id).run()


@dataclass
class _PositionedPeptideFacade:
    start_position: int
    end_position: int
    modified_peptide_sequence: str


class ReverseMzIdentMLPeptide(MzIdentMLPeptide):
    peptide_dict: Dict[str, Any]
    insert_sites: List[Tuple[int, str]]
    deleteion_sites: List[Tuple[int, str]]

    modification_counter: int
    missed_cleavages: Optional[int]

    base_sequence: str
    peptide_sequence: PeptideSequence
    glycosite_candidates: List[int]

    modification_translation_table: Dict[str, Modification]
    enzyme: Optional[Protease]
    mzid_id: Optional[str]

    def __init__(self, peptide_dict, protein: Protein, enzyme=None, constant_modifications=None,
                 modification_translation_table=None, process=True):
        if modification_translation_table is None:
            modification_translation_table = dict()
        if constant_modifications is None:
            constant_modifications = list()

        facade = self.update_sequence(peptide_dict, protein)

        self.peptide_dict = peptide_dict

        self.insert_sites = []
        self.deleteion_sites = []
        self.modification_counter = 0
        self.missed_cleavages = 0

        self.base_sequence = peptide_dict["PeptideSequence"]
        self.peptide_sequence = PeptideSequence(peptide_dict["PeptideSequence"])

        self.glycosite_candidates = n_glycan_sequon_sites(facade, protein)

        self.constant_modifications = constant_modifications
        self.modification_translation_table = modification_translation_table
        self.enzyme = enzyme
        self.mzid_id = peptide_dict.get('id')

        if process:
            self.process()

    def _reflect_position(self, position: int) -> int:
        n = len(self.base_sequence)
        return n - position

    def update_sequence(self, peptide_dict: dict, protein: Protein):
        seq = peptide_dict['PeptideSequence']
        start = protein.protein_sequence[::-1].find(seq)
        n = len(protein)
        # Reflect the coordinates around the reversal point
        end = n - start
        start = end - len(seq)
        seq = protein.protein_sequence[start + 1:end + 1]
        peptide_dict['PeptideSequence'] = seq
        return _PositionedPeptideFacade(start, end, seq)

    def handle_substitutions(self):
        if "SubstitutionModification" in self.peptide_dict:
            subs = self.peptide_dict["SubstitutionModification"]
            for sub in subs:
                pos = sub['location'] - 1
                pos = max(self._reflect_position(pos) - 2, 0)
                replace = Residue(sub["replacementResidue"])
                self.peptide_sequence.substitute(pos, replace)
                self.modification_counter += 1

    def add_modification(self, modification: Modification, position: int):
        if position == -1:
            targets = modification.rule.n_term_targets
            for t in targets:
                if t.position_modifier is not None and t.amino_acid_targets is None:
                    break
            else:
                position += 1
        if position == len(self.peptide_sequence):
            targets = modification.rule.c_term_targets
            for t in targets:
                if t.position_modifier is not None and t.amino_acid_targets is None:
                    break
            else:
                position -= 1
        if position == -1:
            self.peptide_sequence.n_term = modification
        elif position == len(self.peptide_sequence):
            self.peptide_sequence.c_term = modification
        else:
            new_position = max(self._reflect_position(position) - 2, 0)
            self.peptide_sequence.add_modification(new_position, modification)

    def add_insertion(self, site: int, symbol=None):
        site = self._reflect_position(site)
        self.insert_sites.append((site, symbol))

    def add_deletion(self, site: int, symbol):
        site = self._reflect_position(site)
        self.deleteion_sites.append((site, symbol))

    def __repr__(self):
        return f"{self.__class__.__name__}({self.peptide_sequence}, {self.glycosite_candidates})"


class ReversePeptideIdentification(ReverseMzIdentMLPeptide, _EvidenceMixin):
    pass


class ReversePeptideConverter(PeptideConverter):
    def _get_evidence_list(self, peptide_dict: Dict[str, Any]) -> List[Dict[str, Any]]:
        evidence_list = peptide_dict["PeptideEvidenceRef"]
        # Flatten the evidence list if it has extra nesting because of alternative
        # mzid parsing
        if isinstance(evidence_list[0], list):
            evidence_list = [x for sub in evidence_list for x in sub]
        return evidence_list

    def __init__(self, session, hypothesis_id, enzyme=None, constant_modifications=None,
                 modification_translation_table=None, protein_filter=None,
                 peptide_length_range=(5, 60),
                 include_cysteine_n_glycosylation: bool = False,
                 protein_alias_map: Optional[Dict[str, str]]=None):
        super().__init__(
            session, hypothesis_id, enzyme=enzyme, constant_modifications=constant_modifications,
            modification_translation_table=modification_translation_table, protein_filter=protein_filter,
            peptide_length_range=peptide_length_range,
            include_cysteine_n_glycosylation=include_cysteine_n_glycosylation,
            protein_alias_map=protein_alias_map
        )

    def handle_peptide_dict(self, peptide_dict: Dict[str, Any]):
        evidence_list = self._get_evidence_list(peptide_dict)
        for evidence in evidence_list:
            if "skip" in evidence:
                continue
            if evidence.get("isDecoy", False):
                continue

            parent_protein = self.get_protein(evidence)
            if parent_protein is None:
                continue

            peptide_ident = ReversePeptideIdentification(
                peptide_dict,
                parent_protein,
                self.enzyme,
                self.constant_modifications,
                self.modification_translation_table
            )

            score, score_type = peptide_ident.score_type

            start = evidence["start"] - 1
            end = evidence["end"]
            length = len(peptide_ident.base_sequence)
            if not (self.peptide_length_range[0] <= length <= self.peptide_length_range[1]):
                continue

            sequence_copy = peptide_ident.original_sequence()
            found = self.sequence_starts_at(sequence_copy, parent_protein)

            if found != start:
                start = found
                end = start + length

            match = self.pack_peptide(
                peptide_ident, start, end, score, score_type, parent_protein)
            self.add_to_group(match)
            if self.has_occupied_glycosites(match):
                cleared = self.clear_sites(match)
                self.add_to_group(cleared)


class ReverseProteome(Proteome, ProteinReversingMixin):
    def _make_protein_from_state(self, name: str, seq: str, state: dict):
        p = super()._make_protein_from_state(name, seq, state)
        p.sites = []
        self.reverse_protein(p)
        return p

    def _make_peptide_converter(self, session, protein_filter: Set[int], enzyme: Protease, **kwargs):
        return ReversePeptideConverter(
            session,
            self.hypothesis_id,
            enzyme=enzyme,
            constant_modifications=self.constant_modifications,
            modification_translation_table=self.modification_translation_table,
            protein_filter=protein_filter,
            protein_alias_map=self.accession_map, **kwargs)



import operator
import logging

from typing import Callable, List, Optional, TypeVar, Generic
from abc import ABCMeta, abstractmethod, abstractproperty
from collections.abc import Sequence as SequenceABC

try:
    from itertools import imap
except ImportError:
    imap = map

import dill

from six import add_metaclass

from sqlalchemy.orm.session import Session

from glycopeptidepy import HashableGlycanComposition

from .composition_network import CompositionGraph, n_glycan_distance

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


T = TypeVar("T")


@add_metaclass(ABCMeta)
class SearchableMassCollection(Generic[T]):
    def __len__(self):
        return len(self.structures)

    def __iter__(self):
        return iter(self.structures)

    def __getitem__(self, index) -> T:
        return self.structures[index]

    def _convert(self, bundle):
        return bundle

    @abstractproperty
    def lowest_mass(self) -> float:
        raise NotImplementedError()

    def reset(self, **kwargs):
        pass

    @abstractproperty
    def highest_mass(self) -> float:
        raise NotImplementedError()

    def search_mass_ppm(self, mass: float, error_tolerance: float) -> List[T]:
        """Search for the set of all items in :attr:`structures` within `error_tolerance` PPM
        of the queried `mass`.

        Parameters
        ----------
        mass : float
            The neutral mass to search for
        error_tolerance : float, optional
            The range of mass errors (in Parts-Per-Million Error) to allow

        Returns
        -------
        list
            The list of instances which meet the criterion
        """
        tol = mass * error_tolerance
        return self.search_mass(mass, tol)

    @abstractmethod
    def search_mass(self, mass: float, error_tolerance: float=0.1) -> List[T]:
        raise NotImplementedError()

    def __repr__(self):
        size = len(self)
        return "{self.__class__.__name__}({size})".format(self=self, size=size)

    def mark_hit(self, match):
        # no-op to be taken when a structure
        # gets matched.
        return match

    def mark_batch(self):
        # no-op to be taken when a batch of
        # structures is to be searched. This hint
        # helps finalize any intermediary operations.
        pass

SequenceABC.register(SearchableMassCollection)


class MassDatabase(SearchableMassCollection[T]):
    """A quick-to-search database of :class:`HashableGlycanComposition` instances
    stored in memory.

    Implements the Sequence interface, with `__iter__`, `__len__`, and `__getitem__`.

    Attributes
    ----------
    structures : list
        A list of :class:`HashableGlycanComposition` instances, sorted by mass
    """
    def __init__(self, structures, network=None, distance_fn=n_glycan_distance,
                 glycan_composition_type=HashableGlycanComposition, sort=True):
        self.glycan_composition_type = glycan_composition_type
        if not isinstance(structures[0], glycan_composition_type):
            structures = list(map(glycan_composition_type, structures))
        self.structures = structures
        if sort:
            self.structures.sort(key=lambda x: x.mass())
        if network is None:
            self.network = CompositionGraph(self.structures)
            if distance_fn is not None:
                self.network.create_edges(1, distance_fn=distance_fn)
        else:
            self.network = network

    @classmethod
    def from_network(cls, network):
        structures = [node.composition for node in network.nodes]
        return cls(structures, network)

    @property
    def lowest_mass(self):
        return self.structures[0].mass()

    @property
    def highest_mass(self):
        return self.structures[-1].mass()

    @property
    def glycan_composition_network(self):
        return self.network

    def search_binary(self, mass, error_tolerance=1e-6):
        """Search within :attr:`structures` for the index of a structure
        with a mass nearest to `mass`, within `error_tolerance`

        Parameters
        ----------
        mass : float
            The neutral mass to search for
        error_tolerance : float, optional
            The approximate error tolerance to accept

        Returns
        -------
        int
            The index of the structure with the nearest mass
        """
        lo = 0
        n = hi = len(self)

        while hi != lo:
            mid = (hi + lo) // 2
            x = self[mid]
            err = x.mass() - mass
            if abs(err) <= error_tolerance:
                best_index = mid
                best_error = abs(err)
                i = mid - 1
                while i >= 0:
                    x = self.structures[i]
                    err = abs((x.mass() - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i -= 1
                i = mid + 1
                while i < n:
                    x = self.structures[i]
                    err = abs((x.mass() - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i += 1
                return best_index
            elif (hi - lo) == 1:
                best_index = mid
                best_error = err
                i = mid - 1
                while i >= 0:
                    x = self.structures[i]
                    err = abs((x.mass() - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i -= 1
                i = mid + 1
                while i < n:
                    x = self.structures[i]
                    err = abs((x.mass() - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i += 1
                return best_index
            elif err > 0:
                hi = mid
            elif err < 0:
                lo = mid

    def search_mass(self, mass, error_tolerance=0.1):
        """Search for the set of all items in :attr:`structures` within `error_tolerance` Da
        of the queried `mass`.

        Parameters
        ----------
        mass : float
            The neutral mass to search for
        error_tolerance : float, optional
            The range of mass errors (in Daltons) to allow

        Returns
        -------
        list
            The list of items instances which meet the mass criterion
        """
        if len(self) == 0:
            return []
        lo_mass = mass - error_tolerance
        hi_mass = mass + error_tolerance
        lo = self.search_binary(lo_mass)
        hi = self.search_binary(hi_mass) + 1
        return [structure for structure in self[lo:hi] if lo_mass <= structure.mass() <= hi_mass]

    def search_between(self, lower, higher):
        if len(self) == 0:
            return []
        lo = self.search_binary(lower)
        hi = self.search_binary(higher) + 1
        return self.__class__(
            [structure for structure in self[lo:hi] if lower <= structure.mass() <= higher], sort=False)


class MassObject(object):
    __slots__ = ('obj', 'mass')

    def __init__(self, obj, mass):
        self.obj = obj
        self.mass = mass

    def __getattr__(self, attr):
        return getattr(self.obj, attr)

    def __reduce__(self):
        return self.__class__, (self.obj, self.mass)

    def __repr__(self):
        return "MassObject(%r, %r)" % (self.obj, self.mass)

    def __lt__(self, other):
        return self.mass < other.mass

    def __gt__(self, other):
        return self.mass > other.mass

    def __eq__(self, other):
        return abs(self.mass - other.mass) < 1e-3

    def __ne__(self, other):
        return abs(self.mass - other.mass) >= 1e-3


def identity(x):
    return x


class _NeutralMassDatabase(SearchableMassCollection[T]):
    def __init__(self, structures, mass_getter=operator.attrgetter("calculated_mass"), sort=True):
        self.mass_getter = mass_getter
        self.structures = self._prepare(structures, sort)

    def _prepare(self, structures, sort=True):
        temp = []
        for obj in structures:
            mass = self.mass_getter(obj)
            temp.append(MassObject(obj, mass))
        if sort:
            temp.sort()
        return temp

    def __reduce__(self):
        return self.__class__, ([], identity, False), self.__getstate__()

    def __getstate__(self):
        return {
            'mass_getter': dill.dumps(self.mass_getter),
            'structures': self.structures
        }

    def __setstate__(self, state):
        self.mass_getter = dill.loads(state['mass_getter'])
        self.structures = state['structures']

    def __iter__(self):
        for obj in self.structures:
            yield obj.obj

    def __getitem__(self, i):
        if isinstance(i, slice):
            return [mo.obj for mo in self.structures[i]]
        return self.structures[i].obj

    @property
    def lowest_mass(self):
        return self.structures[0].mass

    @property
    def highest_mass(self):
        return self.structures[-1].mass

    def search_binary(self, mass, error_tolerance=1e-6):
        """Search within :attr:`structures` for the index of a structure
        with a mass nearest to `mass`, within `error_tolerance`

        Parameters
        ----------
        mass : float
            The neutral mass to search for
        error_tolerance : float, optional
            The approximate error tolerance to accept

        Returns
        -------
        int
            The index of the structure with the nearest mass
        """
        lo = 0
        n = hi = len(self.structures)
        while hi != lo:
            mid = (hi + lo) // 2
            x = self.structures[mid]
            err = x.mass - mass
            if abs(err) <= error_tolerance:
                best_index = mid
                best_error = err
                i = mid - 1
                while i >= 0:
                    x = self.structures[i]
                    err = abs((x.mass - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i -= 1
                i = mid + 1
                while i < n:
                    x = self.structures[i]
                    err = abs((x.mass - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i += 1
                return best_index
            elif (hi - lo) == 1:
                best_index = mid
                best_error = err
                i = mid - 1
                while i >= 0:
                    x = self.structures[i]
                    err = abs((x.mass - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i -= 1
                i = mid + 1
                while i < n:
                    x = self.structures[i]
                    err = abs((x.mass - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i += 1
                return best_index
            elif err > 0:
                hi = mid
            elif err < 0:
                lo = mid

    def search_mass(self, mass, error_tolerance=0.1):
        """Search for the set of all items in :attr:`structures` within `error_tolerance` Da
        of the queried `mass`.

        Parameters
        ----------
        mass : float
            The neutral mass to search for
        error_tolerance : float, optional
            The range of mass errors (in Daltons) to allow

        Returns
        -------
        list
            The list of instances which meet the criterion
        """
        if len(self) == 0:
            return []
        lo_mass = mass - error_tolerance
        hi_mass = mass + error_tolerance
        lo = self.search_binary(lo_mass)
        hi = self.search_binary(hi_mass) + 1
        return [structure.obj for structure in self.structures[lo:hi] if lo_mass <= structure.mass <= hi_mass]

    def search_between(self, lower, higher):
        lo = self.search_binary(lower)
        hi = self.search_binary(higher) + 1
        return self.__class__(
            [structure.obj for structure in self.structures[lo:hi] if lower <= structure.mass <= higher],
            self.mass_getter, sort=False)

    def _merge_neutral_mass_database(self, other, id_fn=id):
        new = {id_fn(x.obj): x for x in self.structures}
        new.update({id_fn(x.obj): x for x in other.structures})
        structures = list(new.values())
        structures.sort()
        self.structures = structures
        return self

    def _merge_other(self, other, id_fn=id):
        new = {id_fn(x): x for x in self}
        new.update({id_fn(x): x for x in other})
        structures = list(new.values())
        self.structures = self._pack(structures, True)
        return self

    def merge(self, other, id_fn=id):
        if isinstance(other, NeutralMassDatabase):
            return self._merge_neutral_mass_database(other, id_fn)
        else:
            return self._merge_other(other, id_fn)


try:
    from glycresoft._c.database.mass_collection import (
        MassObject, NeutralMassDatabaseImpl as _NeutralMassDatabaseImpl)

    class NeutralMassDatabase(_NeutralMassDatabaseImpl, _NeutralMassDatabase[T]):
        pass
except ImportError:
    NeutralMassDatabase = _NeutralMassDatabase


class ConcatenatedDatabase(SearchableMassCollection[T]):
    def __init__(self, databases):
        self.databases = list(databases)

    def reset(self, **kwargs):
        return [d.reset(**kwargs) for d in self.databases]

    @property
    def lowest_mass(self):
        return min(db.lowest_mass for db in self.databases)

    @property
    def highest_mass(self):
        return max(db.highest_mass for db in self.databases)

    def search_mass(self, mass, error_tolerance=0.1):
        """Search for the set of all items in :attr:`structures` within `error_tolerance` Da
        of the queried `mass`.

        Parameters
        ----------
        mass : float
            The neutral mass to search for
        error_tolerance : float, optional
            The range of mass errors (in Daltons) to allow

        Returns
        -------
        list
            The list of instances which meet the criterion
        """
        hits = []
        for database in self.databases:
            hits += database.search_mass(mass, error_tolerance)
        return hits

    def search_between(self, lower, higher):
        result = []
        for database in self.databases:
            subset = database.search_between(lower, higher)
            if len(subset) > 0:
                result.append(subset)
        return self.__class__(result)

    def add(self, database):
        self.databases.append(database)
        return self

    def __len__(self):
        return sum(map(len, self.databases))


class SearchableMassCollectionWrapper(SearchableMassCollection[T]):
    searchable_mass_collection: SearchableMassCollection

    @property
    def highest_mass(self):
        return self.searchable_mass_collection.highest_mass

    @property
    def lowest_mass(self):
        return self.searchable_mass_collection.lowest_mass

    def reset(self, **kwargs):
        return self.searchable_mass_collection.reset(**kwargs)

    # Fake the disk-backed interface

    @property
    def session(self) -> Session:
        return self.searchable_mass_collection.session

    @property
    def hypothesis_id(self):
        return self.searchable_mass_collection.hypothesis_id

    @property
    def hypothesis(self):
        return self.searchable_mass_collection.hypothesis

    def __len__(self):
        return len(self.searchable_mass_collection)

    @property
    def peptides(self):
        return self.searchable_mass_collection.peptides

    @property
    def proteins(self):
        return self.searchable_mass_collection.proteins

    def query(self, *args, **kwargs):
        return self.searchable_mass_collection.query(*args, **kwargs)


class TransformingMassCollectionAdapter(SearchableMassCollectionWrapper[T]):
    def __init__(self, searchable_mass_collection, transformer):
        self.searchable_mass_collection = searchable_mass_collection
        self.transformer = transformer

    def search_mass_ppm(self, mass: float, error_tolerance: float=1e-5) -> List[T]:
        result = self.searchable_mass_collection.search_mass_ppm(mass, error_tolerance)
        return [self.transformer(r) for r in result]

    def __reduce__(self):
        return self.__class__, (self.searchable_mass_collection, None), self.__getstate__()

    def __getstate__(self):
        return {
            "transformer": dill.dumps(self.transformer)
        }

    def __setstate__(self, state):
        self.transformer = dill.loads(state['transformer'])

    def search_mass(self, mass, error_tolerance):  # pylint: disable=signature-differs
        result = self.searchable_mass_collection.search_mass(mass, error_tolerance)
        return [self.transformer(r) for r in result]

    def __getitem__(self, i):
        return self.transformer(self.searchable_mass_collection[i])

    def __iter__(self):
        return imap(self.transformer, self.searchable_mass_collection)

    def search_between(self, lower, higher):
        return self.__class__(
            self.searchable_mass_collection.search_between(lower, higher), self.transformer)


class MassCollectionProxy(SearchableMassCollectionWrapper[T]):
    _session: Optional[Session]
    session_resolver: Callable[[], Session]

    def __init__(self, resolver, session_resolver=None, hypothesis_id=None, hypothesis_resolver=None):
        self._searchable_mass_collection = None
        self._session = None
        self._hypothesis_id = hypothesis_id
        self._hypothesis = None
        self.resolver = resolver
        self.session_resolver = session_resolver
        self.hypothesis_resolver = hypothesis_resolver

    @property
    def hypothesis_id(self):
        if self._hypothesis_id is not None:
            return self._hypothesis_id
        else:
            logger.info("Loading hypothesis ID from resolver")
            self._hypothesis_id = self.searchable_mass_collection.hypothesis_id
            return self._hypothesis_id

    @property
    def session(self) -> Session:
        if self._session is None and self.session_resolver is not None:
            self._session = self.session_resolver()
        if self._session is not None:
            return self._session
        else:
            logger.info("Loading session from primary resolver")
            self._session = self.searchable_mass_collection.session
            return self._session

    @property
    def hypothesis(self):
        if self._hypothesis is not None:
            return self._hypothesis
        elif self.hypothesis_resolver:
            self._hypothesis = self.hypothesis_resolver()
            return self._hypothesis
        else:
            return super().hypothesis

    def query(self, *args, **kwargs):
        return self.session.query(*args, **kwargs)

    @property
    def searchable_mass_collection(self):
        if self._searchable_mass_collection is None:
            self._searchable_mass_collection = self.resolver()
        return self._searchable_mass_collection

    def search_mass_ppm(self, mass: float, error_tolerance: float=1e-5) -> List[T]:
        result = self.searchable_mass_collection.search_mass_ppm(mass, error_tolerance)
        return result

    def __reduce__(self):
        return self.__class__, (None,), self.__getstate__()

    def __getstate__(self):
        return {
            "resolver": dill.dumps(self.resolver),
            "session_resolver": dill.dumps(self.session_resolver),
            "hypothesis_resolver": dill.dumps(self.hypothesis_resolver),
            "hypothesis_id": self._hypothesis_id
        }

    def _has_resolved(self):
        return self._searchable_mass_collection is not None

    def __repr__(self):
        if self._has_resolved():
            size = len(self)
        else:
            size = "?"
        return "{self.__class__.__name__}({size})".format(self=self, size=size)

    def __setstate__(self, state):
        self.resolver = dill.loads(state['resolver'])
        self.session_resolver = dill.loads(state['session_resolver'])
        self.hypothesis_resolver = dill.loads(state['hypothesis_resolver'])
        self._hypothesis_id = state['hypothesis_id']

    def search_mass(self, mass: float, error_tolerance: float) -> List[T]:  # pylint: disable=signature-differs
        result = self.searchable_mass_collection.search_mass(mass, error_tolerance)
        return result

    def __getitem__(self, i) -> T:
        return self.searchable_mass_collection[i]

    def __iter__(self):
        return iter(self.searchable_mass_collection)

    def search_between(self, lower: float, higher: float) -> List[T]:
        return self.searchable_mass_collection.search_between(lower, higher)
